# Copyright © 2025 UChicago Argonne, LLC All right reserved
# Full license accessible at https://github.com//AdvancedPhotonSource/pty-chi/blob/main/LICENSE

from typing import Union, Literal, Callable, Optional, Sequence, TYPE_CHECKING
import math
import gc

import torch
from torch import Tensor
from torchvision.transforms import GaussianBlur
import numpy as np
from numpy import ndarray

import ptychi.maths as pmath
import ptychi.propagate as propagate
from ptychi.timing.timer_utils import timer

if TYPE_CHECKING:
    from ptychi.api.task import PtychographyTask


_default_complex_dtype = torch.complex64


def get_suggested_object_size(positions_px, probe_shape, extra=0):
    h = np.ceil(positions_px[:, 0].max() - positions_px[:, 0].min()) + probe_shape[0] + extra
    w = np.ceil(positions_px[:, 1].max() - positions_px[:, 1].min()) + probe_shape[1] + extra
    return (int(h), int(w))


def rescale_probe(
    probe: Union[ndarray, Tensor],
    patterns: Union[ndarray, Tensor],
    weights: Optional[Union[ndarray, Tensor]] = None,
) -> None:
    """
    Scale probe so that the sum of intensity matches that of the diffraction patterns.

    Parameters
    ----------
    probe : Tensor
        A (n_modes, h, w) or (n_opr_modes, n_modes, h, w) tensor of the probe.
    patterns : Tensor
        A (n, h, w) tensor of diffraction patterns.
    weights : Tensor, optional
        A (n_points, n_opr_modes) tensor of weights for each OPR mode.

    Returns
    -------
    scaled_probe : Tensor
        The scaled probe.
    """
    propagator = propagate.FourierPropagator()

    probe_tensor = torch.tensor(probe)
    
    if probe_tensor.ndim == 4:
        if probe_tensor.shape[0] == 1 or weights is None:
            probe_tensor = probe_tensor[0]

    if probe_tensor.ndim == 3:
        i_probe = (
            (torch.abs(propagator.propagate_forward(probe_tensor)) ** 2)
            .sum()
            .detach()
            .cpu()
            .numpy()
        )
    else:
        weights = torch.tensor(weights)
        weights = weights.mean(dim=0) 
        probe_corrected = (probe_tensor * weights[:, None, None, None]).sum(0)
        i_probe = (
            (torch.abs(propagator.propagate_forward(probe_corrected)) ** 2)
            .sum()
            .detach()
            .cpu()
            .numpy()
        )

    patterns = to_numpy(patterns)
    i_data = np.sum(np.mean(patterns, axis=0))
    factor = i_data / i_probe
    probe = probe * np.sqrt(factor)
    return probe


def orthogonalize_initial_probe(
    probe: Tensor, 
    secondary_mode_energy: float = 0.02, 
    method: Literal["hermite"] = "hermite"
) -> Tensor:
    """
    Orthogonalize initial probe. 
    
    Parameters
    ----------
    probe : Tensor
        A (n_opr_modes, n_modes, h, w) tensor of the probe. This function only generates
        incoherent modes; OPR modes are kept as they are. Only the first incoherent mode
        of the input probe is used. As such, the rest of the incoherent modes can be
        arbotrarily initialized, but the shape of the input probe should be indicate the
        number of incoherent modes intended to be generated.
    secondary_mode_energy : float, optional
        The energy of the secondary mode relative to the principal mode, which is
        always 1.0.
    method: Literal["hermite"], optional
        The method to use for orthogonalization.
    
    Returns
    -------
    Tensor
        The orthogonalized probe.
    """
    n_modes = probe.shape[1]
    mode_energies = torch.zeros(n_modes)
    mode_energies[1:] = secondary_mode_energy
    mode_energies[0] = 1.0 - mode_energies.sum()
    
    e_total = torch.sum(torch.abs(probe[0, 0]) ** 2)
    mode_energies = mode_energies * e_total
    
    if method == "hermite":
        m = math.ceil(math.sqrt(n_modes)) - 1
        n = math.ceil(n_modes / (m + 1)) - 1
        h = generate_secondary_probe_modes_hermite(probe[0, 0], m, n)
        probe[0, 1:, :, :] = h[1:n_modes, :, :]
    else:
        raise ValueError(f"Unknown orthogonalization method: {method}")
    
    # Normalize.
    for i_mode in range(n_modes):
        probe[0, i_mode] = probe[0, i_mode] * torch.sqrt(mode_energies[i_mode] / torch.sum(torch.abs(probe[0, i_mode] ** 2)))
    return probe


@timer()
def generate_secondary_probe_modes_hermite(probe: Tensor, m: int, n: int) -> Tensor:
    """
    Generate secondary probe modes using Hermite polynomials.
    
    Parameters
    ----------
    probe : Tensor
        A (h, w) tensor of the primary mode of the probe.
    m, n : int
        The orders of the Hermite polynomial.
    
    Returns
    -------
    Tensor
        A ((m + 1) * (n + 1), h, w) tensor of the secondary probe modes.
    """
    x = torch.arange(probe.shape[-1]) - probe.shape[-1] / 2 + 1
    y = torch.arange(probe.shape[-2]) - probe.shape[-2] / 2 + 1
    xx, yy = torch.meshgrid(x, y, indexing="xy")
    
    cenx = torch.sum(xx * torch.abs(probe) ** 2) / torch.sum(torch.abs(probe) ** 2)
    ceny = torch.sum(yy * torch.abs(probe) ** 2) / torch.sum(torch.abs(probe) ** 2)
    varx = torch.sum((xx - cenx) ** 2 * torch.abs(probe) ** 2) / torch.sum(torch.abs(probe) ** 2)
    vary = torch.sum((yy - ceny) ** 2 * torch.abs(probe) ** 2) / torch.sum(torch.abs(probe) ** 2)
    
    counter = 0
    h = torch.empty([(m + 1) * (n + 1), *probe.shape], dtype=probe.dtype)
    for nii in range(n + 1):
        for mii in range(m + 1):
            auxfunc = ((xx - cenx) ** mii) * ((yy - ceny) ** nii) * probe
            if counter > 0:
                auxfunc = auxfunc * torch.exp(-((xx - cenx) ** 2 / (2 * varx)) - ((yy - ceny) ** 2 / (2 * vary)))
            auxfunc = auxfunc / torch.sqrt(torch.sum(torch.abs(auxfunc) ** 2))    

            # Orthogonalize the current mode to the previous ones.
            for ii in range(counter):
                auxfunc = auxfunc - h[ii] * torch.sum(h[ii] * auxfunc.conj(), dim=(-1, -2))
            auxfunc = auxfunc / torch.sqrt(torch.sum(torch.abs(auxfunc) ** 2))
            h[counter] = auxfunc
            counter += 1
    return h


@timer()
def get_probe_renormalization_factor(patterns: Tensor | ndarray) -> float:
    """
    Calculate the renormalization factor that should be applied to the probe
    to match the maximum power of the diffraction patterns.
    
    Parameters
    ----------
    patterns : Tensor | ndarray
        A (n, h, w) buffer of diffraction patterns.
    
    Returns
    -------
    float
        The renormalization factor.
    """
    if isinstance(patterns, Tensor):
        patterns = patterns.detach().cpu().numpy()
    max_power = np.max(np.sum((patterns), axis=(1, 2))) / (patterns[0].size)
    return np.sqrt(1 / max_power)
            

def generate_initial_object(shape: tuple[int, ...], method: Literal["random"] = "random") -> Tensor:
    if method == "random":
        obj_mag = generate_gaussian_random_image(shape, loc=0.98, sigma=0.02, smoothing=3.0)
        obj_mag = obj_mag.clamp(0.0, 1.0)
        obj_phase = generate_gaussian_random_image(shape, loc=0.0, sigma=0.02, smoothing=3.0)
        obj_phase = obj_phase.clamp(-torch.pi, torch.pi)
        obj = obj_mag * torch.exp(1j * obj_phase)
    else:
        raise ValueError(f"Unknown object initialization method: {method}")
    obj = obj.type(get_default_complex_dtype())
    return obj


@timer()
def add_additional_opr_probe_modes_to_probe(
    probe: Tensor, n_opr_modes_to_add: int, normalize: bool = True
) -> Tensor:
    """
    Add additional OPR modes to the probe.
    
    Parameters
    ----------
    probe : Tensor
        A (n_opr_modes, n_modes, h, w) tensor of the probe.
    n_opr_modes_to_add : int
        The number of OPR modes to add.
    normalize : bool, optional
        Whether to normalize the OPR modes using `mnorm` so that the power
        of each mode is the number of pixels in a mode.
    
    Returns
    -------
    Tensor
        A (n_opr_modes + n_opr_modes_to_add, n_modes, h, w) tensor of the probe 
        with additional OPR modes.
    """
    if probe.ndim != 4:
        raise ValueError("probe must be a (n_opr_modes, n_modes, h, w) tensor.")
    n_modes = probe.shape[1]
    opr_modes = torch.empty(
        [n_opr_modes_to_add, n_modes, probe.shape[-2], probe.shape[-1]],
        dtype=get_default_complex_dtype(),
    )
    
    for i in range(n_opr_modes_to_add):
        for j in range(n_modes):
            real = generate_gaussian_random_image(
                probe.shape[-2:], loc=0, sigma=1, smoothing=0
            )
            imag = generate_gaussian_random_image(
                probe.shape[-2:], loc=0, sigma=1, smoothing=0
            )
            opr_mode = real + 1j * imag
            opr_modes[i, j, ...] = opr_mode
    probe = torch.cat([probe, opr_modes], dim=0)
    
    if normalize:
        pnorm = pmath.mnorm(probe, dim=(-2, -1), keepdims=True)
        probe[1:] = probe[1:] / pnorm[1:, :]
    return probe


@timer()
def generate_initial_opr_mode_weights(
    n_points: int, n_opr_modes: int, eigenmode_weight: Optional[float] = None, probe: Optional[Tensor] = None
) -> Tensor:
    """
    Generate initial weights for OPR modes, where the weights of the main OPR mode are set to 1,
    and the weights of eigenmodes are set to 0.

    Parameters
    ----------
    n_points : int
        number of scan points.
    n_opr_modes : int
        number of OPR modes.
    eigenmode_weight : float
        initial weight for eigenmodes.
    probe: Tensor
        The probe. If provided, the weights will be normalized to match the power of the probe.

    Returns
    -------
    weights : Tensor
        a (n_points, n_opr_modes) tensor of weights.
    """
    if eigenmode_weight is None:
        eigenmode_weights = torch.randn([n_points, n_opr_modes - 1]) * 1e-6
    else:
        eigenmode_weights = torch.full([n_points, n_opr_modes - 1], eigenmode_weight)
    weights = torch.cat(
        [torch.ones([n_points, 1]), eigenmode_weights],
        dim=1,
    )
    if probe is not None:
        pnorm = pmath.mnorm(probe, dim=(-2, -1), keepdims=False)
        weights[:, 1:] = weights[:, 1:] / torch.mean(pnorm[1:], dim=1)
    return weights


@timer()
def generate_gaussian_random_image(
    shape: tuple[int, ...], loc: float = 0.9, sigma: float = 0.1, smoothing: float = 3.0
) -> Tensor:
    img = torch.randn(shape, dtype=torch.get_default_dtype()) * sigma + loc
    if smoothing > 0.0:
        img = GaussianBlur(kernel_size=(9, 9), sigma=(3, 3))(img[None, None, :, :])
        img = img[0, 0, ...]
    return img


def to_tensor(data: Union[ndarray, Tensor], device=None, dtype=None) -> Tensor:
    if device is None:
        device = torch.get_default_device()
    if isinstance(data, (np.ndarray, list, tuple)):
        data = torch.tensor(data, device=device)

    if dtype is None:
        if data.dtype.is_complex:
            dtype = get_default_complex_dtype()
        elif not data.dtype.is_complex:
            dtype = torch.get_default_dtype()

    if data.dtype != dtype:
        data = data.type(dtype)
    if str(data.device) != str(device):
        data = data.to(device)
    return data


def to_numpy(data: Union[ndarray, Tensor]) -> ndarray:
    if isinstance(data, Tensor):
        data = data.detach().cpu().numpy()
    return data


def set_default_complex_dtype(dtype):
    global _default_complex_dtype
    _default_complex_dtype = dtype


def get_default_complex_dtype():
    return _default_complex_dtype


def chunked_processing(
    func: Callable,
    common_kwargs: dict,
    chunkable_kwargs: dict,
    iterated_kwargs: dict,
    chunk_size: int = 96,
):
    """
    Parameters
    ----------
    func : callable
        The callable to be executed.
    common_kwargs : dict
        A dictionary of arguments that should stay constant across chunks.
    chunkable_kwargs : dict
        A dictionary of arguments that should be chunked.
    iterated_kwargs : dict
        A dictionary of arguments that should be returned by `func`, then passed to `func`
        for the next chunk. The order of arguments should be the same as the returns of
        `func`.
    chunk_size : int, optional
        The size of each chunk. Default is 96.

    Returns
    -------
    The returns of `func` as if it is executed for the entire data.
    """
    full_batch_size = tuple(chunkable_kwargs.values())[0].shape[0]
    for key, value in tuple(chunkable_kwargs.items())[1:]:
        if value.shape[0] != full_batch_size:
            raise ValueError(
                "All chunkable arguments must have the same batch size, but {} \
                has shape {}.".format(key, value.shape)
            )

    chunks_of_chunkable_args = []
    ind_st = 0
    while ind_st < full_batch_size:
        ind_end = min(ind_st + chunk_size, full_batch_size)
        chunk = {key: value[ind_st:ind_end] for key, value in chunkable_kwargs.items()}
        chunks_of_chunkable_args.append(chunk)
        ind_st = ind_end

    for kwargs_chunk in chunks_of_chunkable_args:
        ret = func(**common_kwargs, **kwargs_chunk, **iterated_kwargs)
        if isinstance(ret, tuple):
            for i, key in enumerate(iterated_kwargs.keys()):
                iterated_kwargs[key] = ret[i]
        else:
            iterated_kwargs[tuple(iterated_kwargs.keys())[0]] = ret
    if len(iterated_kwargs) == 1:
        return tuple(iterated_kwargs.values())[0]
    else:
        return tuple(iterated_kwargs.values())


def calculate_data_size_gb(
    shape: Sequence[int], 
    dtype: torch.dtype | np.dtype
) -> float:
    """
    Calculate the size of the data in GB.
    """
    shape = list(shape)
    return np.prod(shape) * dtype.itemsize / 1024 ** 3


def get_max_batch_size(
    probe_shape: Sequence[int], 
    object_shape: Sequence[int],
    double_precision: bool = False, 
    data_saved_on_device: bool = False, 
    all_data_shape: Sequence[int] = None,
    reconstructor_type: Literal["lsqml"] = "lsqml",
    margin_factor: float = 0.2
) -> int:
    """
    Estimate the maximum batch size that fits in the available device memory.
    
    We estimate the memory usage using an empirical formula:
    ```
    mem = x0 * n_p * batch_size + x1 * n_p + x2 * object_numel
    ```
    where `n_p = n_modes * probe_size ** 2` and the coefficients `x0`, `x1`, `x2` 
    were fit from experimental data.
    
    Parameters
    ----------
    probe_shape : Sequence[int]
        The shape of the 4D probe, expected to be (n_opr_modes, n_modes, h, w).
    object_shape : Sequence[int]
        The shape of the object, expected to be (n_slcies, h, w).
    double_precision : bool, optional
        Whether to use double precision.
    data_saved_on_device : bool, optional
        Whether the raw data is kept on device.
    all_data_shape : Sequence[int], optional
        The shape of the data, expected to be (n_points, h, w).
    reconstructor_type : Literal["lsqml"], optional
        The type of reconstructor. Currently only `lsqml` is supported.
    margin_factor : float, optional
        The fraction of the device memory to be left free.
    
    Returns
    -------
    int
        The suggested batch size.
    """
    if reconstructor_type == "lsqml":
        x0 = 6.26e-8
        x1 = 3.73e-7
        x2 = 1.39e-7
    else:
        raise ValueError(f"Unknown reconstructor type: {reconstructor_type}")
    
    dtype = torch.float64 if double_precision else torch.float32
    n_p = np.prod(list(probe_shape[1:]))
    n_o = np.prod(list(object_shape))
    if data_saved_on_device:
        data_size_gb = calculate_data_size_gb(all_data_shape, dtype)
    else:
        data_size_gb = 0.0
        
    mem_avail = torch.cuda.mem_get_info()[0] * (1 - margin_factor) / 1024 ** 3
    mem_compute = mem_avail - data_size_gb
    batch_size = (mem_compute - x1 * n_p - x2 * n_o) / (x0 * n_p)
    batch_size = batch_size * (8 / dtype.itemsize)
    return max(int(batch_size), 1)


def auto_transfer_to_device(data: Tensor) -> Tensor:
    """Automatically determine the device that the data should be placed on, 
    and transfer the data to that device.
    
    The logic of this function is as follows:
    1. If `torch.get_default_device()` is `cuda`, transfer the data to `cuda`.
    2. If `torch.get_default_device()` is `cpu`, it could either be GPU is unavailable or
       intentionally disabled, OR the current code is executed by DataParallel. 
    2.1. If `torch.cuda.device_count()` is 0, we assume it is the former case, and
         we keep the data as is.
    2.2. If `torch.cuda.device_count()` is not 0, we assume it is the latter case, and
         we transfer the data to `cuda`.
    """
    if torch.get_default_device().type == "cuda":
        return data.cuda()
    else:
        if torch.cuda.device_count() == 0:
            return data
        else:
            return data.cuda()


def clear_memory(task: Optional["PtychographyTask"] = None):
    """Clear the memory of the device used. If a `Task` object is provided,
    it will be deleted and the memory will be released.
    
    Parameters
    ----------
    task : PtychographyTask, optional
        The `Task` object to be deleted.
    """
    if task is not None:
        del task
    gc.collect()
    torch.cuda.empty_cache()
    torch.cuda.ipc_collect()



import argparse

import torch

import ptychi.maths as pmath
import test_utils as tutils


class TestMaths(tutils.BaseTester):
    
    def test_orthgonalize_gs(self):
        torch.manual_seed(123)
        x = torch.rand(2, 3, 4, 5)
        x = pmath.orthogonalize_gs(x, dim=(-2, -1), group_dim=1)
        prod = torch.sum(x[0, 0] * x[0, 1])
        assert prod < 1e-4


    def test_orthgonalize_svd(self):
        torch.manual_seed(123)
        x = torch.rand(2, 4, 5) + 1j * torch.rand(2, 4, 5)
        x = pmath.orthogonalize_svd(x, dim=(-2, -1), group_dim=0)
        prod = torch.sum(x[0] * x[1].conj())
        assert prod.abs() < 1e-4

        x = torch.rand(2, 3, 4, 5) + 1j * torch.rand(2, 3, 4, 5)
        x = pmath.orthogonalize_svd(x, dim=(-2, -1), group_dim=1)
        prod = torch.sum(x[0, 0] * x[0, 1].conj())
        assert prod.abs() < 1e-4

        x = torch.rand(2, 3, 4, 5) + 1j * torch.rand(2, 3, 4, 5)
        x = pmath.orthogonalize_svd(x, dim=(-2, -1), group_dim=0)
        prod = torch.sum(x[0, 0] * x[1, 0].conj())
        assert prod.abs() < 1e-4


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--generate-gold', action='store_true')
    args = parser.parse_args()
    
    tester = TestMaths()
    tester.setup_method(name="", generate_data=False, generate_gold=args.generate_gold, debug=True)
    tester.test_orthgonalize_gs()
    tester.test_orthgonalize_svd()



# Copyright © 2025 UChicago Argonne, LLC All right reserved
# Full license accessible at https://github.com//AdvancedPhotonSource/pty-chi/blob/main/LICENSE

import torch
import torch.nn as nn


class Autoencoder(nn.Module):
    def __init__(
        self, 
        num_in_channels: int = 1, 
        num_levels: int = 3, 
        base_channels: int = 32, 
        use_batchnorm: bool = False, 
        zero_conv: bool = False,
        sigmoid_on_magnitude: bool = True,
        scaled_tanh_on_phase: bool = True
    ):
        """
        Convolutional autoencoder model with adjustable number of levels.

        Parameters
        ----------
        num_in_channels : int
            The number of input channels.
        num_levels : int
            The number of levels in the autoencoder.
        base_channels : int
            The base number of channels. In the encoder part, the number of output channels
            of level `i` is `base_channels * 2 ** i`.
        use_batchnorm : bool
            Whether to use batch normalization.
        zero_conv: bool
            If True, a zero-conv layer, i.e., a conv layer with weights and biases initialized
            to zeros, is added at the end of both decoders.
        sigmoid_on_magnitude: bool
            If True, apply a sigmoid function on the magnitude output to confine the values
            between 0 and 1.
        scaled_tanh_on_phase: bool
            If True, apply a tanh function on the phase output and scale it by pi.
        """
        super(Autoencoder, self).__init__()
        self.num_levels = num_levels
        self.num_in_channels = num_in_channels
        self.base_channels = base_channels
        self.use_batchnorm = use_batchnorm
        self.zero_conv = zero_conv
        self.sigmoid_on_magnitude = sigmoid_on_magnitude
        self.scaled_tanh_on_phase = scaled_tanh_on_phase

        self.build_encoder()
        self.build_magnitude_decoder()
        self.build_phase_decoder()
        
    def build_encoder(self):
        down_blocks = []
        for level in range(self.num_levels):
            down_blocks += self.get_down_block(level)
        self.encoder = nn.Sequential(
            # Appears sequential has similar functionality as TF avoiding need for separate model definition and activ
            *down_blocks
        )
        
    def create_generic_decoder_layers(self) -> list[nn.Module]:
        up_blocks = []
        for level in range(self.num_levels - 1, -1, -1):
            up_blocks += self.get_up_block(level)
        decoder_layers = [
            *up_blocks,
            nn.Conv2d(self.base_channels * 2, 1, 3, stride=1, padding=(1, 1)),
        ]
        return decoder_layers
    
    def create_zero_conv_layer(self, num_in_channels, num_out_channels) -> nn.Module:
        zero_conv = nn.Conv2d(num_in_channels, num_out_channels, kernel_size=1)
        torch.nn.init.zeros_(zero_conv.weight)
        torch.nn.init.zeros_(zero_conv.bias)
        return zero_conv
        
    def build_magnitude_decoder(self):
        decoder = self.create_generic_decoder_layers()
        if self.sigmoid_on_magnitude:
            decoder.append(nn.Sigmoid())
        if self.zero_conv:
            decoder.append(self.create_zero_conv_layer(1, 1))
        self.decoder1 = nn.Sequential(*decoder)
        
    def build_phase_decoder(self):
        decoder = self.create_generic_decoder_layers()
        if self.scaled_tanh_on_phase:
            decoder.append(nn.Tanh())
        if self.zero_conv:
            decoder.append(self.create_zero_conv_layer(1, 1))
        self.decoder2 = nn.Sequential(*decoder)

    def get_down_block(self, level):
        """
        Get a list of layers in a downsampling block.

        Parameters
        ----------
        level : int
            0-based level index.
            
        Returns
        -------
        list[nn.Module]
            List of layers in the downsampling block.
        """
        num_in_channels = int(self.base_channels * 2 ** (level - 1))
        num_out_channels = int(self.base_channels * 2 ** level)
        if level == 0:
            num_in_channels = self.num_in_channels

        blocks = []
        blocks.append(nn.Conv2d(in_channels=num_in_channels, out_channels=num_out_channels,
                      kernel_size=3, stride=1, padding=(1, 1)))
        if self.use_batchnorm:
            blocks.append(nn.BatchNorm2d(num_out_channels))
        blocks.append(nn.ReLU())
        blocks.append(nn.Conv2d(num_out_channels, num_out_channels, 3, stride=1, padding=(1, 1)))
        if self.use_batchnorm:
            blocks.append(nn.BatchNorm2d(num_out_channels))
        blocks.append(nn.ReLU())
        blocks.append(nn.MaxPool2d((2, 2)))

        return blocks

    def get_up_block(self, level):
        """
        Get a list of layers in a upsampling block.
        Parameters
        ----------
        level : int
            0-based level index.
            
        Returns
        -------
        list[nn.Module]
            List of layers in the upsampling block.
        """
        if level == self.num_levels - 1:
            num_in_channels = self.base_channels * 2 ** level
            num_out_channels = self.base_channels * 2 ** level
        elif level == 0:
            num_in_channels = self.base_channels * 2 ** (level + 1)
            num_out_channels = self.base_channels * 2 ** (level + 1)
        else:
            num_in_channels = self.base_channels * 2 ** (level + 1)
            num_out_channels = self.base_channels * 2 ** level
        num_in_channels = int(num_in_channels)
        num_out_channels = int(num_out_channels)

        blocks = []
        blocks.append(nn.Conv2d(num_in_channels, num_out_channels, 3, stride=1, padding=(1, 1)))
        if self.use_batchnorm:
            blocks.append(nn.BatchNorm2d(num_out_channels))
        blocks.append(nn.ReLU())
        blocks.append(nn.Conv2d(num_out_channels, num_out_channels, 3, stride=1, padding=(1, 1)))
        if self.use_batchnorm:
            blocks.append(nn.BatchNorm2d(num_out_channels))
        blocks.append(nn.ReLU())
        blocks.append(nn.Upsample(scale_factor=2, mode='bilinear'))
        return blocks

    def forward(self, x):
        x1 = self.encoder(x)
        amp = self.decoder1(x1)
        ph = self.decoder2(x1)

        if self.scaled_tanh_on_phase:
            ph = ph * torch.pi  # Using tanh activation (-1 to 1) for phase so multiply by pi

        return amp, ph



# Copyright © 2025 UChicago Argonne, LLC All right reserved
# Full license accessible at https://github.com//AdvancedPhotonSource/pty-chi/blob/main/LICENSE

import dataclasses
from dataclasses import field

import ptychi.api.options.base as base
import ptychi.api.options.task as task_options
import ptychi.api.enums as enums


@dataclasses.dataclass
class PIEReconstructorOptions(base.ReconstructorOptions):
    
    def get_reconstructor_type(self) -> enums.Reconstructors:
        return enums.Reconstructors.PIE


@dataclasses.dataclass
class PIEObjectOptions(base.ObjectOptions):
    
    alpha: float = 0.1
    """
    Multiplier for the update to the object, as defined in table 1 of Maiden (2017).
    """


@dataclasses.dataclass
class PIEProbeExperimentalOptions(base.Options):
    sdl_probe_options: base.SynthesisDictLearnProbeOptions = dataclasses.field(default_factory=base.SynthesisDictLearnProbeOptions)


@dataclasses.dataclass
class PIEProbeOptions(base.ProbeOptions):
    
    alpha: float = 0.1
    """
    Multiplier for the update to the probe, as defined in table 1 of Maiden (2017).
    """
    
    experimental: PIEProbeExperimentalOptions = dataclasses.field(default_factory=PIEProbeExperimentalOptions)


@dataclasses.dataclass
class PIEProbePositionOptions(base.ProbePositionOptions):
    pass


@dataclasses.dataclass
class PIEOPRModeWeightsOptions(base.OPRModeWeightsOptions):
    pass


@dataclasses.dataclass
class PIEOptions(task_options.PtychographyTaskOptions):
    
    reconstructor_options: PIEReconstructorOptions = field(default_factory=PIEReconstructorOptions)
    
    object_options: PIEObjectOptions = field(default_factory=PIEObjectOptions)
    
    probe_options: PIEProbeOptions = field(default_factory=PIEProbeOptions)
    
    probe_position_options: PIEProbePositionOptions = field(default_factory=PIEProbePositionOptions)
    
    opr_mode_weight_options: PIEOPRModeWeightsOptions = field(default_factory=PIEOPRModeWeightsOptions)


@dataclasses.dataclass
class EPIEReconstructorOptions(PIEReconstructorOptions):
    
    def get_reconstructor_type(self) -> enums.Reconstructors:
        return enums.Reconstructors.EPIE


@dataclasses.dataclass
class EPIEOptions(PIEOptions):

    reconstructor_options: EPIEReconstructorOptions = field(default_factory=EPIEReconstructorOptions)


@dataclasses.dataclass
class RPIEReconstructorOptions(PIEReconstructorOptions):
    
    def get_reconstructor_type(self) -> enums.Reconstructors:
        return enums.Reconstructors.RPIE


@dataclasses.dataclass
class RPIEOptions(PIEOptions):

    reconstructor_options: RPIEReconstructorOptions = field(default_factory=RPIEReconstructorOptions)


import argparse

import torch
import numpy as np

import ptychi.api as api
from ptychi.api.task import PtychographyTask
from ptychi.utils import get_suggested_object_size, get_default_complex_dtype, generate_initial_opr_mode_weights

import test_utils as tutils


class TestMultislicePtychoLsqmlJointStepSize(tutils.TungstenDataTester):

    @tutils.TungstenDataTester.wrap_recon_tester(name='test_multislice_ptycho_lsqml_joint_step_size')
    def test_multislice_ptycho_lsqml_joint_step_size(self):
        self.setup_ptychi(cpu_only=False)

        data, probe, pixel_size_m, positions_px = self.load_tungsten_data(pos_type='true', additional_opr_modes=3)
        
        options = api.LSQMLOptions()
        options.data_options.data = data
        
        options.object_options.initial_guess = torch.ones([1, *get_suggested_object_size(positions_px, probe.shape[-2:], extra=100)], dtype=get_default_complex_dtype())
        options.object_options.pixel_size_m = pixel_size_m
        options.object_options.optimizable = True
        options.object_options.optimizer = api.Optimizers.SGD
        options.object_options.step_size = 1
        options.object_options.build_preconditioner_with_all_modes = True
        
        options.probe_options.initial_guess = probe
        options.probe_options.optimizable = True
        options.probe_options.optimizer = api.Optimizers.SGD
        options.probe_options.step_size = 1

        options.probe_position_options.position_x_px = positions_px[:, 1]
        options.probe_position_options.position_y_px = positions_px[:, 0]
        options.probe_position_options.optimizable = False
        
        options.opr_mode_weight_options.initial_weights = generate_initial_opr_mode_weights(len(positions_px), probe.shape[0])
        options.opr_mode_weight_options.optimizable = True
        options.opr_mode_weight_options.update_relaxation = 0.1
        
        options.reconstructor_options.batch_size = 96
        options.reconstructor_options.noise_model = api.NoiseModels.GAUSSIAN
        options.reconstructor_options.num_epochs = 8
        options.reconstructor_options.solve_obj_prb_step_size_jointly_for_first_slice_in_multislice = True
        options.reconstructor_options.allow_nondeterministic_algorithms = False

        task = PtychographyTask(options)
        task.run()
        
        recon = task.get_data_to_cpu('object', as_numpy=True)
        return recon[0]
    
    
if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--generate-gold', action='store_true')
    args = parser.parse_args()
    
    tester = TestMultislicePtychoLsqmlJointStepSize()
    tester.setup_method(name="", generate_data=False, generate_gold=args.generate_gold, debug=True)
    tester.test_multislice_ptycho_lsqml_joint_step_size()



# Copyright © 2025 UChicago Argonne, LLC All right reserved
# Full license accessible at https://github.com//AdvancedPhotonSource/pty-chi/blob/main/LICENSE

from typing import Optional, Union, TYPE_CHECKING, Sequence
import dataclasses
from dataclasses import field
import logging
from math import ceil

from numpy import ndarray
from torch import Tensor
import torch
import numpy as np

import ptychi.api.enums as enums
from ptychi.api.options.plan import OptimizationPlan

if TYPE_CHECKING:
    import ptychi.api.options.task as task_options
    from ptychi.api.options.task import PtychographyTaskOptions


logger = logging.getLogger(__name__)


@dataclasses.dataclass
class Options:
    
    def __setattr__(self, name, value):
        # Check if the attribute already exists in the class fields.
        if name not in {f.name for f in dataclasses.fields(self)}:
            raise AttributeError(f"{name} is not a valid field in {self.__class__.__name__}.")
        # If it exists, allow setting the value.
        super().__setattr__(name, value)
        
    def check(self, *args, **kwargs) -> None:
        """Check if options values are valid.
        """
        return


@dataclasses.dataclass
class ParameterOptions(Options):
    optimizable: bool = True
    """
    Whether the parameter is optimizable.
    """

    optimization_plan: OptimizationPlan = dataclasses.field(default_factory=OptimizationPlan)
    """
    Optimization plan for the parameter.
    """

    optimizer: enums.Optimizers = enums.Optimizers.SGD
    """
    Name of the optimizer.
    """

    step_size: float = 1
    """
    Step size of the optimizer. This will be the learning rate `lr` in
    `optimizer_params`.
    """

    optimizer_params: dict = dataclasses.field(default_factory=dict)
    """
    Settings for the optimizer of the parameter. For additional information on
    optimizer parameters, see: https://pytorch.org/docs/stable/optim.html
    """
    
    def check(self, options: "task_options.PtychographyTaskOptions"):
        return super().check(options)

    def get_non_data_fields(self) -> dict:
        d = self.__dict__.copy()
        return d


@dataclasses.dataclass
class FeatureOptions(Options):
    """
    Abstract base class that is inherited by sub-feature dataclasses. This class is used to
    determining if/when a feature is used.
    """

    enabled: bool
    "Turns execution of the feature on and off."

    optimization_plan: OptimizationPlan
    "Schedules when the feature is executed."

    def is_enabled_on_this_epoch(self, current_epoch: int):
        if self.enabled and self.optimization_plan.is_enabled(current_epoch):
            return True
        else:
            return False


@dataclasses.dataclass
class ObjectMultisliceRegularizationOptions(FeatureOptions):
    """Settings for multislice regularization of the object."""

    enabled: bool = False

    optimization_plan: OptimizationPlan = dataclasses.field(default_factory=OptimizationPlan)

    weight: float = 0
    """
    The weight for multislice regularization. Disabled if 0, or if `type != ObjectTypes.MULTISLICE`. 
    When enabled, multislice objects are regularized using cross-slice smoothing.
    """

    unwrap_phase: bool = True
    """Whether to unwrap the phase of the object during multislice regularization."""

    unwrap_image_grad_method: enums.ImageGradientMethods = (
        enums.ImageGradientMethods.FOURIER_DIFFERENTIATION
    )
    """
    The method for calculating the phase gradient during phase unwrapping.
    
        - FOURIER_SHIFT: Use Fourier shift to perform shift.
        - NEAREST: Use nearest neighbor to perform shift.
        - FOURIER_DIFFERENTIATION: Use Fourier differentiation.
    """

    unwrap_image_integration_method: enums.ImageIntegrationMethods = (
        enums.ImageIntegrationMethods.FOURIER
    )
    """
    The method for integrating the phase gradient during phase unwrapping.
    
        - FOURIER: Use Fourier integration as implemented in PtychoShelves.
        - DECONVOLUTION: Deconvolve a ramp filter.
        - DISCRETE: Use cumulative sum.
    """


@dataclasses.dataclass
class ObjectL1NormConstraintOptions(FeatureOptions):
    """Settings for the L1 norm constraint."""

    enabled: bool = False

    optimization_plan: OptimizationPlan = dataclasses.field(default_factory=OptimizationPlan)

    weight: float = 0
    """The weight of the L1 norm constraint. Disabled if equal or less than 0."""
    

@dataclasses.dataclass
class ObjectL2NormConstraintOptions(FeatureOptions):
    """Settings for the L2 norm constraint."""

    enabled: bool = False

    optimization_plan: OptimizationPlan = dataclasses.field(default_factory=OptimizationPlan)

    weight: float = 0
    """The weight of the L2 norm constraint. Disabled if equal or less than 0."""


@dataclasses.dataclass
class ObjectSmoothnessConstraintOptions(FeatureOptions):
    """Settings for smoothing of the magnitude (but not the phase) of the object"""

    enabled: bool = False

    optimization_plan: OptimizationPlan = dataclasses.field(default_factory=OptimizationPlan)

    alpha: float = 0
    """
    The relaxation smoothing constant. This value should be in the range  0 < alpha <= 1/8.

    Smoothing is done by constructing a 3x3 kernel of
    ```
        alpha, alpha,         alpha
        alpha, 1 - 8 * alpha, alpha
        alpha, alpha,         alpha
    ```
    and convolve it with the object magnitude. When `alpha == 1 / 8`, the smoothing power
    is maximal. The value of alpha should not be larger than 1 / 8.
    """


@dataclasses.dataclass
class ObjectTotalVariationOptions(FeatureOptions):
    """Settings for total variation constraint on the object."""

    enabled: bool = False

    optimization_plan: OptimizationPlan = dataclasses.field(default_factory=OptimizationPlan)

    weight: float = 0
    """The weight of the total variation constraint. Disabled if equal or less than 0."""


@dataclasses.dataclass
class RemoveGridArtifactsOptions(FeatureOptions):
    """Settings for grid artifact removal in the object's phase, applied at the end of an epoch"""

    enabled: bool = False

    optimization_plan: OptimizationPlan = dataclasses.field(default_factory=OptimizationPlan)

    period_x_m: float = 1e-7
    """The horizontal period of grid artifacts in meters."""

    period_y_m: float = 1e-7
    """The vertical period of grid artifacts in meters."""

    window_size: int = 5
    """The window size for grid artifact removal in pixels."""

    direction: enums.Directions = enums.Directions.XY
    """The direction of grid artifact removal."""
    

@dataclasses.dataclass
class RemoveObjectProbeAmbiguityOptions(FeatureOptions):
    """Settings for removing the object-probe ambiguity, where the object is scaled by its norm
    so that the mean transmission is kept around 1, and the probe is scaled accordingly.
    """

    enabled: bool = True

    optimization_plan: OptimizationPlan = dataclasses.field(default_factory=lambda: OptimizationPlan(stride=10))
    
    
@dataclasses.dataclass
class SliceSpacingOptions(ParameterOptions):
    
    optimizable: bool = False
    """Whether the slice spacings are optimizable.
    
    Known issue: slice spacing optimization only works with AutodiffPtychography, and we
    have to use `loss.backward(retain_graph=True)` to make it work with AD. This might
    result in growing per-epoch walltime and memory usage. We are working on a better
    solution.
    """
    
    optimization_plan: OptimizationPlan = dataclasses.field(default_factory=OptimizationPlan)
    
    optimizer: enums.Optimizers = enums.Optimizers.SGD
    """The optimizer to use for optimizing the slice spacings."""
    
    step_size: float = 1e-10
    """The step size for the optimizer. As a recommendation, start with 1e-10 for SGD
    optimizer, and 1e-7 for ADAM optimizer.
    """
    
    def check(self, options: "task_options.PtychographyTaskOptions"):
        super().check(options)
        
        if (self.optimizable 
            and options.reconstructor_options.get_reconstructor_type() != enums.Reconstructors.AD_PTYCHO
        ):
            raise ValueError("Slice spacing optimization is only supported for AD Ptychography.")
    
    

@dataclasses.dataclass
class ObjectOptions(ParameterOptions):
    initial_guess: Union[ndarray, Tensor] = None
    """A (h, w) complex tensor of the object initial guess."""

    slice_spacings_m: Optional[ndarray] = None
    """Slice spacings in meters. This should be provided if the object is multislice.
    
    If the slice spacings need to be optimized, set `slice_spacing_options.optimizable` to `True`.
    In that case, the slice spacings provided here are supposed to be the initial guess.
    """
    
    slice_spacing_options: SliceSpacingOptions = field(default_factory=SliceSpacingOptions)

    pixel_size_m: float = 1.0
    """The pixel size in meters. When pixel size is non-square, this should be the width (x)
    of the pixel size."""
    
    pixel_size_aspect_ratio: float = 1.0
    """The aspect ratio of the pixel size, defined as width (x) / height (y).
    """

    l1_norm_constraint: ObjectL1NormConstraintOptions = field(
        default_factory=ObjectL1NormConstraintOptions
    )
    
    l2_norm_constraint: ObjectL2NormConstraintOptions = field(
        default_factory=ObjectL2NormConstraintOptions
    )

    smoothness_constraint: ObjectSmoothnessConstraintOptions = field(
        default_factory=ObjectSmoothnessConstraintOptions
    )

    total_variation: ObjectTotalVariationOptions = field(
        default_factory=ObjectTotalVariationOptions
    )

    remove_grid_artifacts: RemoveGridArtifactsOptions = field(
        default_factory=RemoveGridArtifactsOptions
    )

    multislice_regularization: ObjectMultisliceRegularizationOptions = field(
        default_factory=ObjectMultisliceRegularizationOptions
    )

    patch_interpolation_method: enums.PatchInterpolationMethods = (
        enums.PatchInterpolationMethods.FOURIER
    )
    """
    Selects the interpolation method used for extracting and updating 
    patches of the object IF patch extraction/placement is done using 
    the object's methods `extract_patches_function` or 
    `place_patches_function`.
    """
    
    remove_object_probe_ambiguity: RemoveObjectProbeAmbiguityOptions = field(
        default_factory=RemoveObjectProbeAmbiguityOptions
    )
    
    build_preconditioner_with_all_modes: bool = False
    """If True, the probe illumination map used for the preconditioner is 
    built using the sum of intensities of all probe modes. This may help address
    some issues if some probe modes contain highly localized high-intensity anomalies,
    if the selected reconstructor uses preconditioner to regularize object updates.
    However, it might lead to slower convergence speed.
    """
    
    determine_position_origin_coords_by: enums.ObjectPosOriginCoordsMethods = enums.ObjectPosOriginCoordsMethods.SUPPORT
    """The method to determine the pixel coordinates of the object that corresponds 
    to the origin of the probe positions. 
    
    Probe positions are given as a list of coordinates that can be either positive
    or negative and have arbitrary offsets, while the object buffer is a discrete
    tensor where the pixel indices are 0-based and the origin is at the top left corner.
    The position origin coordinates are used to determine how the given probe positions
    are mapped to the object buffer: the origin of the probe positions (0, 0) is mapped
    to the pixel indices given by the position origin coordinates, and as such, the
    pixel indices of all probe positions are calculated as
    ```
    positions_pxind = positions + position_origin_coords
    ```
    
    - `POSITIONS`: the origin coordinates are determined as 
      `buffer_center - (positions.max() + positions.min()) / 2`. This puts the mid-point
      of the position range at the center of the buffer. It is more adaptive; however,
      in the case that one initializes a reconstruction with a previously reconstructed object
      and corrected probe positions, the center coordinates are not necessarily the same, 
      which can cause the positions to mismatch between both reconstructions. 
      
    - `SUPPORT`: the origin coordinates are determined as the center of the support of the 
      object. This is helpful to keep the center coordinates consistent between consecutive
      reconstructions, but the probe positions given should (at least approximately) zero-centered,
      i.e., `-postitions.min() ~ positions.max()` to prevent out-of-bound errors.
      
    - `SPECIFIED`: the origin coordinates are specified by the user. To make this setting effective,
      `position_origin_coords` should be specified.
    """
    
    position_origin_coords: Optional[ndarray] = None
    """The user-specified origin coordinates of the object. To make this setting effective,
    `determine_position_origin_coords_by` should be set to `SPECIFIED`. 
    
    Probe positions are given as a list of coordinates that can be either positive
    or negative and have arbitrary offsets, while the object buffer is a discrete
    tensor where the pixel indices are 0-based and the origin is at the top left corner.
    The position origin coordinates are used to determine how the given probe positions
    are mapped to the object buffer: the origin of the probe positions (0, 0) is mapped
    to the pixel indices given by the position origin coordinates, and as such, the
    pixel indices of all probe positions are calculated as
    ```
    positions_pxind = positions + position_origin_coords
    ```
    """

    def get_non_data_fields(self) -> dict:
        d = super().get_non_data_fields()
        del d["initial_guess"]
        return d
    
    def check(self, options: "PtychographyTaskOptions"):
        super().check(options)
        pos_y = options.probe_position_options.position_y_px
        pos_x = options.probe_position_options.position_x_px
        probe_shape = options.probe_options.initial_guess.shape[-2:]
        obj_shape = options.object_options.initial_guess.shape[-2:]
        min_size = [
            int(ceil((pos_y.max() - pos_y.min() + probe_shape[-2]).item())) + 2,
            int(ceil((pos_x.max() - pos_x.min() + probe_shape[-1]).item())) + 2,
        ]
        if any([min_size[i] > obj_shape[i] for i in range(2)]):
            logging.warning(
                f"An object tensor with a lateral size of at least {min_size} is "
                "required to avoid padding when extracting/placing patches, but the provided "
                f"object size is {list(options.object_options.initial_guess.shape[-2:])}."
            )
        if self.determine_position_origin_coords_by == enums.ObjectPosOriginCoordsMethods.SUPPORT:
            buffer_center = torch.tensor(
                [np.round(x / 2) + 0.5 for x in obj_shape]
            )
            if (
                pos_y.max() + buffer_center[0] + probe_shape[-2] // 2 > obj_shape[-2]
                or pos_y.min() + buffer_center[0] - probe_shape[-2] // 2 < 0
                or pos_x.max() + buffer_center[1] + probe_shape[-1] // 2 > obj_shape[-1]
                or pos_x.min() + buffer_center[1] - probe_shape[-1] // 2 < 0
            ):
                logging.warning(
                    "`object_options.determine_center_coords_by` is set to `SUPPORT`. This assumes "
                    "that the probe positions are approximately zero-centered, i.e., "
                    "`-pos_y.min() ~ pos_y.max()` and `-pos_x.min() ~ pos_x.max()`. "
                    "However, the given probe positions will cause the reconstructor to access pixels "
                    "out of the object support. Please provide probe positions that are approximately "
                    "zero-centered, or set `object_options.determine_center_coords_by` to `POSITIONS`."
                )
        if self.determine_position_origin_coords_by == enums.ObjectPosOriginCoordsMethods.SPECIFIED:
            if self.position_origin_coords is None:
                raise ValueError("`object_options.center_coords` should be specified when "
                                 "`object_options.determine_center_coords_by` is set to "
                                 "`SPECIFIED`.")
        if self.position_origin_coords is not None:
            if self.determine_position_origin_coords_by != enums.ObjectPosOriginCoordsMethods.SPECIFIED:
                logging.warning(
                    "`object_options.center_coords` will be disregarded when "
                    "`object_options.determine_center_coords_by` is not set to "
                    "`SPECIFIED`."
                )
                
        if self.smoothness_constraint.enabled:
            if self.smoothness_constraint.alpha > 1.0 / 8 or self.smoothness_constraint.alpha < 0:
                raise ValueError(
                    f"smoothness_constraint.alpha = {self.smoothness_constraint.alpha} is out of range [0, 1/8]."
                )


@dataclasses.dataclass
class ProbePowerConstraintOptions(FeatureOptions):
    """
    Settings for scaling the probe and object intensity.
    """

    enabled: bool = False

    optimization_plan: OptimizationPlan = dataclasses.field(default_factory=OptimizationPlan)

    probe_power: float = 0.0
    """
    The target probe power. The probe and object intensity will be scaled such that 
    the power of the far-field probe is `probe_power`.
    """


@dataclasses.dataclass
class ProbeOrthogonalizeIncoherentModesOptions(FeatureOptions):
    """
    Settings for orthogonalizing incoherent probe modes.
    """

    enabled: bool = True

    optimization_plan: OptimizationPlan = dataclasses.field(default_factory=OptimizationPlan)

    method: enums.OrthogonalizationMethods = enums.OrthogonalizationMethods.SVD
    """The method to use for incoherent_mode orthogonalization."""


@dataclasses.dataclass
class ProbeOrthogonalizeOPRModesOptions(FeatureOptions):
    """
    Settings for orthogonalizing OPR modes.
    """

    enabled: bool = True

    optimization_plan: OptimizationPlan = dataclasses.field(default_factory=OptimizationPlan)


@dataclasses.dataclass
class ProbeSupportConstraintOptions(FeatureOptions):
    """
    Settings for probe shrinkwrapping, where small values are set to 0.
    """
        
    enabled: bool = False

    optimization_plan: OptimizationPlan = dataclasses.field(default_factory=OptimizationPlan)

    threshold: float = 0.005
    """
    The threshold for the probe support constraint. The value of a pixel (x, y) is set to 0
    if `p(x, y) < [max(blur(p)) * `threshold`](x, y)`.
    """


@dataclasses.dataclass
class ProbeCenterConstraintOptions(FeatureOptions):
    """
    Settings for constraining the probe's center of mass to the center of the probe array.
    """

    enabled: bool = False

    optimization_plan: OptimizationPlan = dataclasses.field(default_factory=OptimizationPlan)


@dataclasses.dataclass
class ProbeOptions(ParameterOptions):
    """
    The probe configuration.

    The first OPR mode of all incoherent modes are always optimized aslong as
    `optimizable` is `True`. In addition to thtat, eigenmodes (of the first
    incoherent mode) are optimized when:

    - The probe has multiple OPR modes;
    - `OPRModeWeightsConfig` is given.
    """

    initial_guess: Union[ndarray, Tensor] = None
    """A (n_opr_modes, n_modes, h, w) complex tensor of the probe initial guess."""

    power_constraint: ProbePowerConstraintOptions = field(
        default_factory=ProbePowerConstraintOptions
    )

    orthogonalize_incoherent_modes: ProbeOrthogonalizeIncoherentModesOptions = field(
        default_factory=ProbeOrthogonalizeIncoherentModesOptions
    )

    orthogonalize_opr_modes: ProbeOrthogonalizeOPRModesOptions = field(
        default_factory=ProbeOrthogonalizeOPRModesOptions
    )

    support_constraint: ProbeSupportConstraintOptions = field(
        default_factory=ProbeSupportConstraintOptions
    )

    center_constraint: ProbeCenterConstraintOptions = field(
        default_factory=ProbeCenterConstraintOptions
    )

    eigenmode_update_relaxation: float = 1.0
    """
    A separate step size for eigenmode update.
    """

    def check(self, options: "task_options.PtychographyTaskOptions"):
        super().check(options)
        if not (self.initial_guess is not None and self.initial_guess.ndim == 4):
            raise ValueError("Probe initial_guess must be a (n_opr_modes, n_modes, h, w) tensor.")
        if self.power_constraint.enabled and options.object_options.remove_object_probe_ambiguity.enabled:
            logger.warning(
                "`ObjectOptions.remove_object_probe_ambiguity` and `ProbeOptions.power_constraint` "
                "are both enabled, which may lead to unexpected results."
            )

    def get_non_data_fields(self) -> dict:
        d = super().get_non_data_fields()
        del d["initial_guess"]
        return d

@dataclasses.dataclass
class SynthesisDictLearnProbeOptions(Options):
    
    d_mat: Union[ndarray, Tensor] = None
    """The synthesis sparse dictionary matrix; contains the basis functions 
    that will be used to represent the probe via the sparse code weights."""
    
    d_mat_conj_transpose: Union[ndarray, Tensor] = None
    """Conjugate transpose of the synthesis sparse dictionary matrix."""
    
    d_mat_pinv: Union[ndarray, Tensor] = None
    """Moore-Penrose pseudoinverse of the synthesis sparse dictionary matrix."""
    
    probe_sparse_code: Union[ndarray, Tensor] = None
    """Sparse code weights vector."""
    
    probe_sparse_code_nnz: float = None
    """Number of non-zeros we will keep when enforcing sparsity constraint on
    the sparse code weights vector probe_sparse_code."""
    
    enabled: bool = False

@dataclasses.dataclass
class PositionCorrectionOptions(Options):
    """Options used for specifying the position correction function."""

    correction_type: enums.PositionCorrectionTypes = enums.PositionCorrectionTypes.GRADIENT
    """Type of algorithm used to calculate the position correction update."""
    
    differentiation_method: enums.ImageGradientMethods = enums.ImageGradientMethods.FOURIER_DIFFERENTIATION
    """The method for calculating the gradient of the object. Only used when `correction_type` 
    is `GRADIENT`. `"FOURIER_DIFFERENTIATION"` is usually the fastest, but it might be less
    stable when the object is noisy or non-smooth, under which circumstance `"GAUSSIAN"` or
    `"FOURIER_SHIFT"` may offer better stability. `"NEAREST"` is not recommended.
    """

    cross_correlation_scale: int = 20000
    """The upsampling factor of the cross-correlation in real space."""

    cross_correlation_real_space_width: float = 0.01
    """The width of the cross-correlation in real-space"""

    cross_correlation_probe_threshold: float = 0.1
    """The probe intensity threshold used to calculate the probe mask."""
    
    slice_for_correction: int = None
    """The object slice for which the position correction is calculated. If None, the middle slice
    is chosen.
    """
    
    clip_update_magnitude_by_mad: bool = True
    """If True, the update magnitude is eventually clipped by 10 times the mean absolute deviation (MAD)
    of the updates. When `update_magnitude_limit` is set, the limit will be set to the smaller of them,
    i.e., `min(update_magnitude_limit, 10 * MAD)`.
    """
    
    update_magnitude_limit: Optional[float] = 0.1
    """The maximum allowed magnitude of position update in each axis. Updates larger than this value 
    are clipped. Set to None or inf to disable the constraint. When `clip_update_magnitude_by_mad` is
    `True`, the actual limit will be set to the smaller of `update_magnitude_limit` and `10 * MAD`.
    """
    

@dataclasses.dataclass
class PositionAffineTransformConstraintOptions(FeatureOptions):
    """Settings for imposing an affine transformation constraint on the probe positions.
    """

    enabled: bool = False

    optimization_plan: OptimizationPlan = dataclasses.field(default_factory=OptimizationPlan)
    
    degrees_of_freedom: Sequence[enums.AffineDegreesOfFreedom] = (
        enums.AffineDegreesOfFreedom.ROTATION,
        enums.AffineDegreesOfFreedom.SCALE,
        enums.AffineDegreesOfFreedom.SHEAR,
        enums.AffineDegreesOfFreedom.ASYMMETRY,
    )
    """The degrees of freedom to include in the affine transformation."""
    
    position_weight_update_interval: int = 10
    """The number of epochs between position weight updates.
    """
    
    apply_constraint: bool = True
    """Constraint is applied to probe positions only when this is `True`. When `False`,
    probe position weights and affine transformation matrix are still computed and
    stored in the `ProbePositions` object so that they can be logged and analyzed 
    externally, but the positions are not altered.
    """
    
    max_expected_error: float = 1.0
    """The maximum expected position error, given in pixels. Note that this is different
    from `update_magnitude_limit`, and is only used in the estimation of friction in
    affine transformation constraint.
    """
    
    def is_position_weight_update_enabled_on_this_epoch(self, current_epoch: int):
        if not self.enabled:
            return False
        if (current_epoch - self.optimization_plan.start) % self.position_weight_update_interval == 0:
            return True
        else:
            return False


@dataclasses.dataclass
class ProbePositionMagnitudeLimitOptions(FeatureOptions):
    """Settings for imposing a magnitude limit on the probe position update."""

    enabled: bool = False

    optimization_plan: OptimizationPlan = dataclasses.field(default_factory=OptimizationPlan)

    limit: Optional[float] = 0


@dataclasses.dataclass
class ProbePositionOptions(ParameterOptions):
    optimizable: bool = False
    
    step_size: float = 0.3
    """The step size for probe position update."""
    
    position_x_px: Union[ndarray, Tensor] = None
    """The x position in pixel."""

    position_y_px: Union[ndarray, Tensor] = None
    """The y position in pixel."""

    constrain_position_mean: bool = False
    """
    Whether to subtract the mean from positions after updating positions.
    """

    correction_options: PositionCorrectionOptions = dataclasses.field(
        default_factory=PositionCorrectionOptions
    )
    """
    Detailed options for position correction.
    """
    
    affine_transform_constraint: PositionAffineTransformConstraintOptions = dataclasses.field(
        default_factory=PositionAffineTransformConstraintOptions
    )
    """When enabled, an affine transformation from initial positions to current positions
    is fit, and positions deviating from the expected positions given by the affine
    transformation are penalized.
    """

    def get_non_data_fields(self) -> dict:
        d = super().get_non_data_fields()
        del d["position_x_px"]
        del d["position_y_px"]
        return d
        
    def check(self, options: "task_options.PtychographyTaskOptions"):
        super().check(options)
        if self.correction_options.update_magnitude_limit == 0:
            raise ValueError(
                "`probe_position_options.correction_options.update_magnitude_limit` is "
                "set to 0. This will prevent the optimizer from performing any updates, but "
                "will at the same time produce unstability. If you want to disable "
                "position correction, set `optimizable` to `False`. To disable update "
                "magnitude limit, set `update_magnitude_limit` to None or inf."
            )


@dataclasses.dataclass
class OPRModeWeightsSmoothingOptions(FeatureOptions):
    """Settings for smoothing OPR mode weights."""

    enabled: bool = False

    optimization_plan: OptimizationPlan = dataclasses.field(default_factory=OptimizationPlan)
    
    method: Optional[enums.OPRWeightSmoothingMethods] = None
    """
    The method for smoothing OPR mode weights. 
    
    MEDIAN: applying a median filter to the weights of each mode. 
    
    POLYNOMIAL: fit the weights of each mode with a polynomial of selected degree.
    """

    polynomial_degree: int = 4
    """
    The degree of the polynomial used for smoothing OPR mode weights.
    """


@dataclasses.dataclass
class OPRModeWeightsOptions(ParameterOptions):
    initial_weights: Union[ndarray] = None
    """
    The initial weight(s) of the eigenmode(s). Acceptable values include the following:
    - a (n_scan_points, n_opr_modes) array of initial weights for every point.
    - a (n_opr_modes,) array that gives the weights of each OPR mode. These weights
        will be duplicated for every point.
    """
    
    optimizable: bool = False
    """
    The master switch of optimizability of OPR mode weights. This option must be set
    to True for either `optimize_eigenmode_weights` or `optimize_intensity_variation`
    to take effect.
    """

    optimize_eigenmode_weights: bool = True
    """
    Whether to optimize eigenmode weights, i.e., the weights of the second and
    following OPR modes.

    At least one of `optimize_eigenmode_weights` and `optimize_intensity_variation`
    should be set to `True` if `optimizable` is `True`.
    """

    optimize_intensity_variation: bool = False
    """
    Whether to optimize intensity variation, i.e., the weight of the first OPR mode.

    At least one of `optimize_eigenmode_weights` and `optimize_intensity_variation`
    should be set to `True` if `optimizable` is `True`.
    """

    smoothing: OPRModeWeightsSmoothingOptions = dataclasses.field(
        default_factory=OPRModeWeightsSmoothingOptions
    )

    update_relaxation: float = 1.0
    """
    A separate step size for eigenmode weight update.
    """

    def check(self, options: "task_options.PtychographyTaskOptions"):
        super().check(options)
        if self.optimizable:
            if not (self.optimize_intensity_variation or self.optimize_eigenmode_weights):
                raise ValueError(
                    "When OPRModeWeights is optimizable, at least 1 of "
                    "optimize_intensity_variation and optimize_eigenmode_weights "
                    "should be set to True."
                )
        n_opr_modes_in_probe = options.probe_options.initial_guess.shape[0]
        if n_opr_modes_in_probe > 1:
            if self.initial_weights is None:
                raise ValueError(
                    f"You have {n_opr_modes_in_probe} OPR modes in the probe initial guess, "
                    "but initial OPR weights are not provided."
                )
            elif self.initial_weights.shape[-1] != n_opr_modes_in_probe:
                raise ValueError(
                    f"You have {n_opr_modes_in_probe} OPR modes in the probe initial guess, "
                    f"but the number of modes in your provided OPR weights is {self.initial_weights.shape[-1]}."
                )
        else:
            if self.initial_weights is None:
                logging.info(
                    "Unspecified OPR weight initial guess will be automatically populated with 1s."
                )
            elif self.initial_weights.shape[-1] != n_opr_modes_in_probe:
                raise ValueError(
                    f"You have {n_opr_modes_in_probe} OPR modes in the probe initial guess, "
                    f"but the number of modes in your provided OPR weights is {self.initial_weights.shape[-1]}."
                )
        if self.initial_weights is not None and self.optimizable:
            logging.warning(
                "The default value of OPRModeWeightsOptions has been changed to False. "
                "You have provided initial OPR weights, but optimizable is set to False. "
                "Is this intended?"            
            )

    def get_non_data_fields(self) -> dict:
        d = super().get_non_data_fields()
        del d["initial_weights"]
        return d
    
    
@dataclasses.dataclass
class ForwardModelOptions(Options):
    low_memory_mode: bool = False
    """If True, forward propagation of ptychography will be done using less vectorized code.
    This reduces the speed, but also lowers memory usage.
    """
    
    pad_for_shift: Optional[int] = 0
    """If not None, the image is padded with border values by this amount before shifting."""


@dataclasses.dataclass
class ReconstructorOptions(Options):
    # This should be superseded by CorrectionPlan in ParameterConfig when it is there.
    num_epochs: int = 100
    """The number of epochs to run."""

    batch_size: int = 100
    """The number of data to process in each minibatch."""

    batching_mode: enums.BatchingModes = enums.BatchingModes.RANDOM
    """
    The batching mode to use. 
    
    - `enums.BatchingModes.RANDOM`: load a random set of data in each minibatch.
    - `enums.BatchingModes.COMPACT`: load a spatially close cluster of data in each minibatch.
      This is equivalent to the "compact" mode in PtychoShelves.
    - `enums.BatchingModes.UNIFORM`: load a random set of data in each minibatch, but the
      indices across batches are manipulated so that points in each batch are more uniformly
      spread out in the scan space. This is equivalent to the "sparse" mode in PtychoShelves.
    """

    compact_mode_update_clustering: bool = False
    """
    If True, clusters are updated after each probe position update when `batching_mode` is
    `COMPACT`.
    """

    compact_mode_update_clustering_stride: int = 1
    """
    The number of epochs between updating clusters when `batching_mode` is `COMPACT` and
    `compact_mode_update_clustering` is `True`.
    """

    default_device: enums.Devices = enums.Devices.GPU
    """The default device to use for computation."""

    default_dtype: enums.Dtypes = enums.Dtypes.FLOAT32
    """The default data type to use for computation."""
    
    use_double_precision_for_fft: bool = False
    """If True, use double precision for critical FFT operations. When set to `True`,
    this option overrides `default_dtype`: even if `default_dtype` is set to `FLOAT32`,
    the FFTs will still be performed using double precision. If `False`,
    the FFTs will be performed using the precision specified by `default_dtype`.
    """

    allow_nondeterministic_algorithms: bool = True
    """If True, allow nondeterministic algorithms to be used. Non-deterministic algorithms
    include `scatter_add_` and `scatter_`. They can be faster, but may produce larger
    run-to-run variations.
    """

    random_seed: Optional[int] = None
    """The random seed to use for reproducibility. If None, no seed will be set."""

    displayed_loss_function: Optional[enums.LossFunctions] = enums.LossFunctions.MSE_SQRT
    """
    The function that computes the displayed cost. Different from the `loss_function`
    argument in some reconstructors, this function is only used for cost displaying
    and is not involved in the reconstruction math.
    """

    forward_model_options: ForwardModelOptions = dataclasses.field(
        default_factory=ForwardModelOptions
    )
    

    def get_reconstructor_type(self) -> enums.Reconstructors:
        return enums.Reconstructors.base


@dataclasses.dataclass
class TaskOptions(Options):
    pass



# Copyright © 2025 UChicago Argonne, LLC All right reserved
# Full license accessible at https://github.com//AdvancedPhotonSource/pty-chi/blob/main/LICENSE

import dataclasses
from dataclasses import field

import ptychi.api.options.base as base
from .data import PtychographyDataOptions


@dataclasses.dataclass
class PtychographyTaskOptions(base.TaskOptions):

    data_options: PtychographyDataOptions = field(default_factory=PtychographyDataOptions)

    reconstructor_options: base.ReconstructorOptions = field(default_factory=base.ReconstructorOptions)

    object_options: base.ObjectOptions = field(default_factory=base.ObjectOptions)

    probe_options: base.ProbeOptions = field(default_factory=base.ProbeOptions)

    probe_position_options: base.ProbePositionOptions = field(default_factory=base.ProbePositionOptions)

    opr_mode_weight_options: base.OPRModeWeightsOptions = field(default_factory=base.OPRModeWeightsOptions)

    def check(self, *args, **kwargs):
        super().check(*args, **kwargs)
        for options in (
            self.data_options,
            self.reconstructor_options,
            self.object_options,
            self.probe_options,
            self.probe_position_options,
            self.opr_mode_weight_options,
        ):
            options.check(self)



import argparse
import pytest

import torch

import ptychi.api as api
from ptychi.api.task import PtychographyTask
from ptychi.utils import get_suggested_object_size, get_default_complex_dtype, generate_initial_opr_mode_weights

import test_utils as tutils


class Test2dPtychoDIP(tutils.TungstenDataTester):
    """Deep image prior testers.
    
    Due to the known run-to-run stochasticity of the DIP, no comparison with reference data is done.
    These testers are only used to test if the end-to-end pipeline is working.
    """

    @tutils.TungstenDataTester.wrap_recon_tester(name='test_2d_ptycho_dip', run_comparison=False)
    def test_2d_ptycho_dip(self):
        self.setup_ptychi(cpu_only=False)
        
        # For now, use high tolerance to avoid failure due to stochasticity.
        self.atol = 0.5
        
        data, probe, pixel_size_m, positions_px = self.load_tungsten_data(additional_opr_modes=0)
        
        options = api.AutodiffPtychographyOptions()
        options.data_options.data = data
        
        options.object_options.initial_guess = torch.ones([1, *get_suggested_object_size(positions_px, probe.shape[-2:], extra=100)], dtype=get_default_complex_dtype())
        options.object_options.experimental.deep_image_prior_options.enabled = True
        # options.object_options.experimental.deep_image_prior_options.model = api.enums.DIPModels.UNET
        # options.object_options.experimental.deep_image_prior_options.model_params = {
        #     "num_in_channels": 32,
        #     "num_out_channels": 2,
        #     "initialize": False,
        #     "skip_connections": False,
        # }
        # options.object_options.experimental.deep_image_prior_options.constrain_object_outside_network = True
        options.object_options.experimental.deep_image_prior_options.model = api.enums.DIPModels.AUTOENCODER
        options.object_options.experimental.deep_image_prior_options.model_params = {
            "num_in_channels": 32,
            "num_levels": 4,
            "base_channels": 32,
            "use_batchnorm": True,
        }
        options.object_options.experimental.deep_image_prior_options.constrain_object_outside_network = False
        options.object_options.pixel_size_m = pixel_size_m
        options.object_options.optimizable = True
        options.object_options.optimizer = api.Optimizers.ADAM
        options.object_options.step_size = 1e-5
        
        options.probe_options.initial_guess = probe
        options.probe_options.optimizable = True
        options.probe_options.optimizer = api.Optimizers.SGD
        options.probe_options.step_size = 0.1

        options.probe_position_options.position_x_px = positions_px[:, 1]
        options.probe_position_options.position_y_px = positions_px[:, 0]
        options.probe_position_options.optimizable = False
        
        options.reconstructor_options.batch_size = 100
        options.reconstructor_options.num_epochs = 16
        options.reconstructor_options.allow_nondeterministic_algorithms = False
        
        task = PtychographyTask(options)
        task.run()
        
        recon = task.get_data_to_cpu('object', as_numpy=True)[0, 250:500, 250:500]
        return recon
    
    @tutils.TungstenDataTester.wrap_recon_tester(name='test_2d_ptycho_dip_l2', run_comparison=False)
    def test_2d_ptycho_dip_l2(self):
        self.setup_ptychi(cpu_only=False)
        
        # For now, use high tolerance to avoid failure due to stochasticity.
        self.atol = 0.5
        
        data, probe, pixel_size_m, positions_px = self.load_tungsten_data(additional_opr_modes=0)
        
        options = api.AutodiffPtychographyOptions()
        options.data_options.data = data
        
        options.object_options.initial_guess = torch.ones([1, *get_suggested_object_size(positions_px, probe.shape[-2:], extra=100)], dtype=get_default_complex_dtype())
        options.object_options.experimental.deep_image_prior_options.enabled = True
        options.object_options.experimental.deep_image_prior_options.model = api.enums.DIPModels.AUTOENCODER
        options.object_options.experimental.deep_image_prior_options.model_params = {
            "num_in_channels": 32,
            "num_levels": 4,
            "base_channels": 32,
            "use_batchnorm": True,
        }
        options.object_options.experimental.deep_image_prior_options.constrain_object_outside_network = False
        options.object_options.pixel_size_m = pixel_size_m
        options.object_options.optimizable = True
        options.object_options.optimizer = api.Optimizers.ADAM
        options.object_options.step_size = 1e-5
        options.object_options.l2_norm_constraint.enabled = True
        options.object_options.l2_norm_constraint.weight = 1e-2
        
        options.probe_options.initial_guess = probe
        options.probe_options.optimizable = True
        options.probe_options.optimizer = api.Optimizers.SGD
        options.probe_options.step_size = 0.1

        options.probe_position_options.position_x_px = positions_px[:, 1]
        options.probe_position_options.position_y_px = positions_px[:, 0]
        options.probe_position_options.optimizable = False
        
        options.reconstructor_options.batch_size = 100
        options.reconstructor_options.num_epochs = 16
        options.reconstructor_options.allow_nondeterministic_algorithms = False
        
        task = PtychographyTask(options)
        task.run()
        
        recon = task.get_data_to_cpu('object', as_numpy=True)[0, 250:500, 250:500]
        return recon
    
    @tutils.TungstenDataTester.wrap_recon_tester(name='test_2d_ptycho_dip_probe', run_comparison=False)
    def test_2d_ptycho_dip_probe(self):
        self.setup_ptychi(cpu_only=False)
        
        # For now, use high tolerance to avoid failure due to stochasticity.
        self.atol = 0.5
        
        data, probe, pixel_size_m, positions_px = self.load_tungsten_data(additional_opr_modes=0)
        
        options = api.AutodiffPtychographyOptions()
        options.data_options.data = data
        
        options.object_options.initial_guess = torch.ones([1, *get_suggested_object_size(positions_px, probe.shape[-2:], extra=100)], dtype=get_default_complex_dtype())
        options.object_options.pixel_size_m = pixel_size_m
        options.object_options.optimizable = True
        options.object_options.optimizer = api.Optimizers.ADAM
        options.object_options.step_size = 1e-1
        
        options.probe_options.initial_guess = probe
        options.probe_options.experimental.deep_image_prior_options.enabled = True
        options.probe_options.experimental.deep_image_prior_options.model = api.enums.DIPModels.AUTOENCODER
        options.probe_options.experimental.deep_image_prior_options.model_params = {
            "num_in_channels": 32,
            "num_levels": 4,
            "base_channels": 32,
            "use_batchnorm": True,
            "zero_conv": True,
            "sigmoid_on_magnitude": False,
            "scaled_tanh_on_phase": False
        }
        options.probe_options.experimental.deep_image_prior_options.residual_generation = True
        options.probe_options.experimental.deep_image_prior_options.constrain_object_outside_network = False
        options.probe_options.optimizable = True
        options.probe_options.optimizer = api.Optimizers.ADAM
        options.probe_options.step_size = 1e-4

        options.probe_position_options.position_x_px = positions_px[:, 1]
        options.probe_position_options.position_y_px = positions_px[:, 0]
        options.probe_position_options.optimizable = False
        
        options.reconstructor_options.batch_size = 100
        options.reconstructor_options.num_epochs = 16
        options.reconstructor_options.allow_nondeterministic_algorithms = False
        
        task = PtychographyTask(options)
        task.run()

        recon = task.get_data_to_cpu('object', as_numpy=True)[0, 250:500, 250:500]
        return recon


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--generate-gold', action='store_true')
    args = parser.parse_args()

    tester = Test2dPtychoDIP()
    tester.setup_method(name="", generate_data=False, generate_gold=args.generate_gold, debug=True)
    tester.test_2d_ptycho_dip()
    tester.test_2d_ptycho_dip_l2()
    tester.test_2d_ptycho_dip_probe()



# Copyright © 2025 UChicago Argonne, LLC All right reserved
# Full license accessible at https://github.com//AdvancedPhotonSource/pty-chi/blob/main/LICENSE

# ruff: noqa: F403

from .data import *
from .ad_ptychography import *
from .lsqml import *
from .pie import *
from .plan import *
from .dm import *
from .bh import *



# Copyright © 2025 UChicago Argonne, LLC All right reserved
# Full license accessible at https://github.com//AdvancedPhotonSource/pty-chi/blob/main/LICENSE

from typing import Optional, TYPE_CHECKING

import torch
from torch.utils.data import Dataset
from torch import Tensor

from ptychi.reconstructors.base import (
    AnalyticalIterativePtychographyReconstructor,
)
from ptychi.metrics import MSELossOfSqrt
from ptychi.timing.timer_utils import timer
import ptychi.image_proc as ip

if TYPE_CHECKING:
    import ptychi.api as api
    import ptychi.data_structures.parameter_group as pg


class PIEReconstructor(AnalyticalIterativePtychographyReconstructor):
    """
    The ptychographic iterative engine (PIE), as described in:

    Andrew Maiden, Daniel Johnson, and Peng Li, "Further improvements to the
    ptychographical iterative engine," Optica 4, 736-745 (2017)

    Object and probe updates are calculated using the formulas in table 1 of
    Maiden (2017).

    The `step_size` parameter is equivalent to gamma in Eq. 22 of Maiden (2017)
    when `optimizer == SGD`.
    """

    parameter_group: "pg.PlanarPtychographyParameterGroup"

    def __init__(
        self,
        parameter_group: "pg.PlanarPtychographyParameterGroup",
        dataset: Dataset,
        options: Optional["api.options.pie.PIEReconstructorOptions"] = None,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(
            parameter_group=parameter_group,
            dataset=dataset,
            options=options,
            *args,
            **kwargs,
        )

    def build_loss_tracker(self):
        if self.displayed_loss_function is None:
            self.displayed_loss_function = MSELossOfSqrt()
        return super().build_loss_tracker()

    def check_inputs(self, *args, **kwargs):
        if self.parameter_group.object.is_multislice:
            raise NotImplementedError("EPIEReconstructor only supports 2D objects.")
        for var in self.parameter_group.get_optimizable_parameters():
            if "lr" not in var.optimizer_params.keys():
                raise ValueError(
                    "Optimizable parameter {} must have 'lr' in optimizer_params.".format(var.name)
                )

    @timer()
    def run_minibatch(self, input_data, y_true, *args, **kwargs):
        self.parameter_group.probe.initialize_grad()
        (delta_o, delta_p_i, delta_pos), y_pred = self.compute_updates(
            *input_data, y_true, self.dataset.valid_pixel_mask
        )
        self.apply_updates(delta_o, delta_p_i, delta_pos)
        self.loss_tracker.update_batch_loss_with_metric_function(y_pred, y_true)

    @timer()
    def compute_updates(
        self, indices: torch.Tensor, y_true: torch.Tensor, valid_pixel_mask: torch.Tensor
    ) -> tuple[torch.Tensor, ...]:
        """
        Calculates the updates of the whole object, the probe, and other parameters.
        This function is called in self.update_step_module.forward.
        """
        object_ = self.parameter_group.object
        probe = self.parameter_group.probe
        probe_positions = self.parameter_group.probe_positions
        opr_mode_weights = self.parameter_group.opr_mode_weights

        indices = indices.cpu()
        positions = probe_positions.tensor[indices]

        y = self.forward_model.forward(indices)
        obj_patches = self.forward_model.intermediate_variables["obj_patches"]
        psi = self.forward_model.intermediate_variables["psi"]
        psi_far = self.forward_model.intermediate_variables["psi_far"]
        unique_probes = self.forward_model.intermediate_variables.shifted_unique_probes[0]

        psi_prime = self.replace_propagated_exit_wave_magnitude(psi_far, y_true)
        # Do not swap magnitude for bad pixels.
        psi_prime = torch.where(
            valid_pixel_mask.repeat(psi_prime.shape[0], probe.n_modes, 1, 1), psi_prime, psi_far
        )
        psi_prime = self.forward_model.free_space_propagator.propagate_backward(psi_prime)

        delta_o = None
        if object_.optimization_enabled(self.current_epoch):
            step_weight = self.calculate_object_step_weight(unique_probes)
            delta_o_patches = step_weight * (psi_prime - psi)
            delta_o_patches = delta_o_patches.sum(1, keepdim=True)
            delta_o = ip.place_patches_integer(
                torch.zeros_like(object_.get_slice(0)),
                positions.round().int() + object_.pos_origin_coords,
                delta_o_patches[:, 0],
                op="add",
            )
            # Add slice dimension.
            delta_o = delta_o.unsqueeze(0)

        delta_pos = None
        if probe_positions.optimization_enabled(self.current_epoch) and object_.optimizable:
            delta_pos = torch.zeros_like(probe_positions.data)
            delta_pos[indices] = probe_positions.position_correction.get_update(
                psi_prime - psi,
                obj_patches,
                delta_o_patches,
                self.forward_model.intermediate_variables.shifted_unique_probes[0],
                object_.optimizer_params["lr"],
            )

        delta_p_i = None
        if probe.optimization_enabled(self.current_epoch) and self.parameter_group.probe.representation == "sparse_code":
            rc = psi_prime.shape[-1] * psi_prime.shape[-2]
            n_scpm = psi_prime.shape[-3]
            n_pos = psi_prime.shape[-4]
            
            psi_prime_vec = torch.reshape(psi_prime, (n_pos, n_scpm, rc))
            
            probe_vec = torch.reshape(self.parameter_group.probe.data[0, ...], (n_scpm , rc))
            
            obj_patches_vec = torch.reshape(obj_patches, (n_pos, 1, rc ))
            
            conj_obj_patches = torch.conj(obj_patches_vec)
            abs2_obj_patches = torch.abs(obj_patches_vec) ** 2
            
            z = torch.sum(abs2_obj_patches, dim = 0)
            z_max = torch.max(z)
            w = 0.9 * (z_max - z)
            
            sum_spos_conjT_s_psi = torch.sum(conj_obj_patches * psi_prime_vec, 0)
            sum_spos_conjT_s_psi = torch.swapaxes(sum_spos_conjT_s_psi, 0, 1)
            
            w_phi =  torch.swapaxes(w * probe_vec, 0, 1)
            z_plus_w = torch.swapaxes(z + w, 0, 1)
            
            numer = self.parameter_group.probe.dictionary_matrix_H @ (sum_spos_conjT_s_psi + w_phi)
            denom = (self.parameter_group.probe.dictionary_matrix_H @ (z_plus_w * self.parameter_group.probe.dictionary_matrix))
            
            sparse_code = torch.linalg.solve(denom, numer)
            
            # Enforce sparsity constraint on sparse code
            abs_sparse_code = torch.abs(sparse_code)
            sparse_code_sorted = torch.sort(abs_sparse_code, dim=0, descending=True)
            
            sel = sparse_code_sorted[0][self.parameter_group.probe.probe_sparse_code_nnz, :]
            
            sparse_code = sparse_code * (abs_sparse_code >= sel)
   
            # Update sparse code in probe object
            self.parameter_group.probe.set_sparse_code(sparse_code)
        else:
            step_weight = self.calculate_probe_step_weight(obj_patches)
            delta_p_i = step_weight * (psi_prime - psi)  # get delta p at each position
            delta_p_i = self.adjoint_shift_probe_update_direction(indices, delta_p_i, first_mode_only=True)

        # Calculate and apply opr mode updates
        if self.parameter_group.opr_mode_weights.optimization_enabled(self.current_epoch):
            opr_mode_weights.update_variable_probe(
                probe,
                indices,
                psi_prime - psi,
                delta_p_i,
                delta_p_i.mean(0),
                obj_patches,
                self.current_epoch,
                probe_mode_index=0,
            )

        return (delta_o, delta_p_i, delta_pos), y

    @timer()
    def calculate_object_step_weight(self, p: Tensor):
        """
        Calculate the weight for the object update step.

        Parameters
        ----------
        p : Tensor
            A (batch_size, n_modes, h, w) tensor giving the first OPR mode of the probe.

        Returns
        -------
        Tensor
            A (batch_size, n_modes, h, w) tensor giving the weight for the object update step.
        """
        numerator = p.abs() * p.conj()
        denominator = p.abs().sum(1, keepdim=True).max() * (
            p.abs() ** 2 + self.parameter_group.object.options.alpha * (p.abs() ** 2).sum(1, keepdim=True).max()
        )
        step_weight = numerator / denominator
        return step_weight

    @timer()
    def calculate_probe_step_weight(self, obj_patches: Tensor):
        """
        Calculate the weight for the probe update step.

        Parameters
        ----------
        obj_patches : Tensor
            A (batch_size, n_slices, h, w) tensor giving the object patches.

        Returns
        -------
        Tensor
        """
        # Just take the first slice.
        obj_patches = obj_patches[:, 0]

        obj_max = (torch.abs(obj_patches) ** 2).max(-1).values.max(-1).values.view(-1, 1, 1)
        numerator = obj_patches.abs() * obj_patches.conj()
        denominator = obj_max * (
            obj_patches.abs() ** 2 + self.parameter_group.probe.options.alpha * obj_max
        )
        step_weight = numerator / denominator
        return step_weight

    @timer()
    def apply_updates(self, delta_o, delta_p_i, delta_pos, *args, **kwargs):
        """
        Apply updates to optimizable parameters given the updates calculated by self.compute_updates.

        Parameters
        ----------
        delta_o : Tensor
            A (h, w, 2) tensor of object update vector.
        delta_p_i : Tensor
            A (n_patches, n_opr_modes, n_modes, h, w, 2) tensor of probe update vector.
        delta_pos : Tensor
            A (n_positions, 2) tensor of probe position vectors.
        """
        object_ = self.parameter_group.object
        probe_positions = self.parameter_group.probe_positions

        if delta_o is not None:
            object_.set_grad(-delta_o)
            object_.optimizer.step()

        if delta_p_i is not None:
            mode_slicer = self.parameter_group.probe._get_probe_mode_slicer(None)
            self.parameter_group.probe.set_grad(-delta_p_i.mean(0), slicer=(0, mode_slicer))
            self.parameter_group.probe.optimizer.step()

        if delta_pos is not None:
            probe_positions.set_grad(-delta_pos)
            probe_positions.step_optimizer()


class EPIEReconstructor(PIEReconstructor):
    """
    The extended ptychographic iterative engine (ePIE), as described in:

    Andrew Maiden, Daniel Johnson, and Peng Li, "Further improvements to the
    ptychographical iterative engine," Optica 4, 736-745 (2017)

    Object and probe updates are calculated using the formulas in table 1 of
    Maiden (2017).

    The `step_size` parameter is equivalent to gamma in Eq. 22 of Maiden (2017)
    when `optimizer == SGD`.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @timer()
    def calculate_object_step_weight(self, p: Tensor):
        p_max = (torch.abs(p) ** 2).sum(1).max()
        step_weight = self.parameter_group.object.options.alpha * p.conj() / p_max
        return step_weight

    @timer()
    def calculate_probe_step_weight(self, obj_patches: Tensor):
        # Just take the first slice.
        obj_patches = obj_patches[:, 0]

        obj_max = (torch.abs(obj_patches) ** 2).max(-1).values.max(-1).values.view(-1, 1, 1)
        step_weight = self.parameter_group.probe.options.alpha * obj_patches.conj() / obj_max
        step_weight = step_weight[:, None]
        return step_weight


class RPIEReconstructor(PIEReconstructor):
    """
    The regularized ptychographic iterative engine (rPIE), as described in:

    Andrew Maiden, Daniel Johnson, and Peng Li, "Further improvements to the
    ptychographical iterative engine," Optica 4, 736-745 (2017)

    Object and probe updates are calculated using the formulas in table 1 of
    Maiden (2017).

    The `step_size` parameter is equivalent to gamma in Eq. 22 of Maiden (2017)
    when `optimizer == SGD`.

    To get the momentum-accelerated PIE (mPIE), use `optimizer == SGD` and use
    the optimizer settings `{'momentum': eta, 'nesterov': True}` where `eta` is
    the constant used in  Eq. 19 of Maiden (2017).
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @timer()
    def calculate_object_step_weight(self, p: Tensor):
        p_max = (torch.abs(p) ** 2).sum(1).max()
        step_weight = p.conj() / (
            (1 - self.parameter_group.object.options.alpha) * (torch.abs(p) ** 2)
            + self.parameter_group.object.options.alpha * p_max
        )
        return step_weight

    @timer()
    def calculate_probe_step_weight(self, obj_patches: Tensor):
        # Just take the first slice.
        obj_patches = obj_patches[:, 0]

        obj_max = (torch.abs(obj_patches) ** 2).max(-1).values.max(-1).values.view(-1, 1, 1)
        step_weight = obj_patches.conj() / (
            (1 - self.parameter_group.probe.options.alpha) * (torch.abs(obj_patches) ** 2)
            + self.parameter_group.probe.options.alpha * obj_max
        )
        step_weight = step_weight[:, None]
        return step_weight



# Copyright © 2025 UChicago Argonne, LLC All right reserved
# Full license accessible at https://github.com//AdvancedPhotonSource/pty-chi/blob/main/LICENSE

from typing import Literal, Union, overload
from types import TracebackType
import random
import logging
import os

import torch
import numpy as np
from torch import Tensor
from numpy import ndarray

import ptychi.api as api
import ptychi.data_structures.object as object
import ptychi.data_structures.opr_mode_weights as oprweights
import ptychi.data_structures.probe as probe
import ptychi.data_structures.probe_positions as probepos
import ptychi.data_structures.parameter_group as paramgrp
import ptychi.maps as maps
from ptychi.io_handles import PtychographyDataset
from ptychi.reconstructors.base import Reconstructor
from ptychi.utils import to_tensor
import ptychi.utils as utils
import ptychi.maths as pmath
from ptychi.timing import timer_utils
import ptychi.movies as movies

logger = logging.getLogger(__name__)


class Task:
    def __init__(self, options: api.options.base.TaskOptions, *args, **kwargs) -> None:
        pass

    def __enter__(self) -> "Task":
        return self

    @overload
    def __exit__(self, exception_type: None, exception_value: None, traceback: None) -> None: ...

    @overload
    def __exit__(
        self,
        exception_type: type[BaseException],
        exception_value: BaseException,
        traceback: TracebackType,
    ) -> None: ...

    def __exit__(
        self,
        exception_type: type[BaseException] | None,
        exception_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        torch.cuda.empty_cache()


class PtychographyTask(Task):
    def __init__(self, options: api.options.task.PtychographyTaskOptions, *args, **kwargs) -> None:
        super().__init__(options, *args, **kwargs)
        self.options = options
        self.data_options = options.data_options
        self.object_options = options.object_options
        self.probe_options = options.probe_options
        self.position_options = options.probe_position_options
        self.opr_mode_weight_options = options.opr_mode_weight_options
        self.reconstructor_options = options.reconstructor_options

        self.dataset = None
        self.object = None
        self.probe = None
        self.probe_positions = None
        self.opr_mode_weights = None
        self.reconstructor: Reconstructor | None = None

        self.check_options()
        self.build()
        
    def check_options(self):
        self.options.check()

    def build(self):
        self.build_random_seed()
        self.build_default_device()
        self.build_default_dtype()
        self.build_data()
        self.build_object()
        self.build_probe()
        self.build_probe_positions()
        self.build_opr_mode_weights()
        self.build_reconstructor()

    def build_random_seed(self):
        if self.reconstructor_options.random_seed is not None:
            torch.manual_seed(self.reconstructor_options.random_seed)
            np.random.seed(self.reconstructor_options.random_seed)
            random.seed(self.reconstructor_options.random_seed)
        pmath.set_allow_nondeterministic_algorithms(self.reconstructor_options.allow_nondeterministic_algorithms)

    def build_default_device(self):
        torch.set_default_device(maps.get_device_by_enum(self.reconstructor_options.default_device))
        if torch.cuda.device_count() > 0:
            cuda_visible_devices_str = "(unset)"
            if "CUDA_VISIBLE_DEVICES" in os.environ.keys():
                cuda_visible_devices_str = os.environ["CUDA_VISIBLE_DEVICES"]
            logger.info(
                "Using device: {} (CUDA_VISIBLE_DEVICES=\"{}\")".format(
                    [torch.cuda.get_device_name(i) for i in range(torch.cuda.device_count())],
                    cuda_visible_devices_str,
                )
            )
        else:
            logger.info("Using device: {}".format(torch.get_default_device()))

    def build_default_dtype(self):
        torch.set_default_dtype(maps.get_dtype_by_enum(self.reconstructor_options.default_dtype))
        utils.set_default_complex_dtype(
            maps.get_complex_dtype_by_enum(self.reconstructor_options.default_dtype)
        )
        pmath.set_use_double_precision_for_fft(
            self.reconstructor_options.use_double_precision_for_fft
        )

    def build_data(self):
        if self.data_options.free_space_propagation_distance_m < np.inf and self.data_options.fft_shift:
            logger.warning(
                "It seems that you are reconstructing near-field data with FFT-shifted diffraction data. "
                "Is this intended? If not, set `data_options.fft_shift=False`."
            )
        self.dataset = PtychographyDataset(
            self.data_options.data, 
            wavelength_m=self.data_options.wavelength_m,
            free_space_propagation_distance_m=self.data_options.free_space_propagation_distance_m,
            fft_shift=self.data_options.fft_shift,
            save_data_on_device=self.data_options.save_data_on_device,
            valid_pixel_mask=self.data_options.valid_pixel_mask
        )

    def build_object(self):
        data = to_tensor(self.object_options.initial_guess)
        kwargs = {
            "data": data,
            "options": self.object_options,
        }
        if (
            isinstance(self.object_options, api.options.AutodiffPtychographyObjectOptions)
        ) and (
            self.object_options.experimental.deep_image_prior_options.enabled
        ):
            self.object = object.DIPPlanarObject(**kwargs)
        else:
            self.object = object.PlanarObject(**kwargs)

    def build_probe(self):
        data = to_tensor(self.probe_options.initial_guess)
        kwargs = {
            "data": data,
            "options": self.probe_options,
        }
        if (
            isinstance(self.probe_options, api.options.AutodiffPtychographyProbeOptions)
        ) and (
            self.probe_options.experimental.deep_image_prior_options.enabled
        ):
            self.probe = probe.DIPProbe(**kwargs)
        elif (
            isinstance(self.probe_options, api.options.PIEProbeOptions)
        ) and (
            self.probe_options.experimental.sdl_probe_options.enabled
        ):
            self.probe = probe.SynthesisDictLearnProbe(**kwargs)
        else:
            self.probe = probe.Probe(**kwargs)

    def build_probe_positions(self):
        pos_y = to_tensor(self.position_options.position_y_px)
        pos_x = to_tensor(self.position_options.position_x_px)
        data = torch.stack([pos_y, pos_x], dim=1)
        self.probe_positions = probepos.ProbePositions(data=data, options=self.position_options)

    def build_opr_mode_weights(self):
        if self.opr_mode_weight_options.initial_weights is None:
            initial_weights = torch.ones([self.data_options.data.shape[0], 1])
        else:
            initial_weights = to_tensor(self.opr_mode_weight_options.initial_weights)
        if initial_weights.ndim == 1:
            # If a 1D array is given, expand it to all scan points.
            initial_weights = initial_weights.unsqueeze(0).repeat(
                len(self.position_options.position_x_px), 1
            )
        self.opr_mode_weights = oprweights.OPRModeWeights(
            data=initial_weights, options=self.opr_mode_weight_options
        )

    def build_reconstructor(self):
        par_group = paramgrp.PlanarPtychographyParameterGroup(
            object=self.object,
            probe=self.probe,
            probe_positions=self.probe_positions,
            opr_mode_weights=self.opr_mode_weights,
        )

        reconstructor_class = maps.get_reconstructor_by_enum(
            self.reconstructor_options.get_reconstructor_type()
        )

        reconstructor_kwargs = {
            "parameter_group": par_group,
            "dataset": self.dataset,
            "options": self.reconstructor_options,
        }

        self.reconstructor = reconstructor_class(**reconstructor_kwargs)
        self.reconstructor.build()

    def run(self, n_epochs: int = None):
        """
        Run reconstruction either for `n_epochs` (if given), or for the number of epochs given
        in the options. The internal states of the Task object persists when this function
        finishes. To run more epochs continuing from the last run, call this function again.

        Parameters
        ----------
        n_epochs : int, optional
            The number of epochs to run. If None, use the number of epochs specified in the
            option object.
        """
        if movies.MOVIES_INSTALLED and self.reconstructor.current_epoch == 0:
            movies.api.reset_movie_builders()
        timer_utils.clear_timer_globals()
        self.reconstructor.run(n_epochs=n_epochs)

    def get_data(
        self, name: Literal["object", "probe", "probe_positions", "opr_mode_weights"]
    ) -> Tensor:
        """Get a detached copy of the data of the given name.

        Parameters
        ----------
        name : Literal["object", "probe", "probe_positions", "opr_mode_weights"]
            The name of the data to get.

        Returns
        -------
        Tensor
            The data of the given name.
        """
        return getattr(self, name).data.detach()

    def get_data_to_cpu(
        self,
        name: Literal["object", "probe", "probe_positions", "opr_mode_weights"],
        as_numpy: bool = False,
    ) -> Union[Tensor, ndarray]:
        data = self.get_data(name).cpu()
        if as_numpy:
            data = data.numpy()
        return data
    
    def get_probe_positions_y(self, as_numpy: bool = False) -> Union[Tensor, ndarray]:
        data = self.probe_positions.data[:, 0].detach()
        if as_numpy:
            data = data.cpu().numpy()
        return data

    def get_probe_positions_x(self, as_numpy: bool = False) -> Union[Tensor, ndarray]:
        data = self.probe_positions.data[:, 1].detach()
        if as_numpy:
            data = data.cpu().numpy()
        return data
    
    def copy_data_from_task(
        self, 
        task: "PtychographyTask",
        params_to_copy: tuple[str, ...] = ("object", "probe", "probe_positions", "opr_mode_weights")
    ) -> None:
        """Copy data of reconstruction parameters from another task object.

        Parameters
        ----------
        task : PtychographyTask
            The task object to copy from.
        params_to_copy : tuple[str, ...], optional
            The parameters to copy. By default, copy all parameters.
        """
        with torch.no_grad():
            for param in params_to_copy:
                if param == "object":
                    self.reconstructor.parameter_group.object.set_data(
                        task.get_data("object")
                    )
                elif param == "probe":
                    self.reconstructor.parameter_group.probe.set_data(
                        task.get_data("probe")
                    )
                elif param == "probe_positions":
                    self.reconstructor.parameter_group.probe_positions.set_data(
                        task.get_data("probe_positions")
                    )
                elif param == "opr_mode_weights":
                    self.reconstructor.parameter_group.opr_mode_weights.set_data(
                        task.get_data("opr_mode_weights")
                    )
                else:
                    raise ValueError(f"Invalid parameter name: {param}")

    def __exit__(self, exc_type, exc_value, exc_tb):
        del self.object
        del self.probe
        del self.probe_positions
        del self.opr_mode_weights
        del self.reconstructor
        del self.dataset

        super().__exit__(exc_type, exc_value, exc_tb)



import argparse

import torch

import ptychi.api as api
from ptychi.api.task import PtychographyTask
from ptychi.utils import (
    get_suggested_object_size,
    get_default_complex_dtype,
    rescale_probe,
    generate_initial_opr_mode_weights,
)
from ptychi.maths import orthogonalize_gs

import test_utils as tutils


class Test2dPtychoEpieOPR(tutils.TungstenDataTester):
    @tutils.TungstenDataTester.wrap_recon_tester(name="test_2d_ptycho_epie_opr")
    def test_2d_ptycho_epie_opr(self):
        self.setup_ptychi(cpu_only=False)

        data, probe, pixel_size_m, positions_px = self.load_tungsten_data(additional_opr_modes=3)

        options = api.EPIEOptions()
        options.data_options.data = data

        options.object_options.initial_guess = torch.ones(
            [1, *get_suggested_object_size(positions_px, probe.shape[-2:], extra=100)],
            dtype=get_default_complex_dtype(),
        )
        options.object_options.pixel_size_m = pixel_size_m
        options.object_options.optimizable = True
        options.object_options.optimizer = api.Optimizers.SGD
        options.object_options.step_size = 0.1
        options.object_options.alpha = 1

        options.probe_options.initial_guess = probe
        options.probe_options.optimizable = True
        options.probe_options.optimizer = api.Optimizers.SGD
        options.probe_options.step_size = 0.1
        options.probe_options.alpha = 1
        options.probe_options.eigenmode_update_relaxation = 0.1

        options.probe_position_options.position_x_px = positions_px[:, 1]
        options.probe_position_options.position_y_px = positions_px[:, 0]
        options.probe_position_options.optimizable = False

        options.opr_mode_weight_options.initial_weights = generate_initial_opr_mode_weights(
            len(positions_px), probe.shape[0]
        )
        options.opr_mode_weight_options.optimizable = True
        options.opr_mode_weight_options.update_relaxation = 0.1

        options.reconstructor_options.batch_size = 96
        options.reconstructor_options.num_epochs = 32
        options.reconstructor_options.allow_nondeterministic_algorithms = False

        task = PtychographyTask(options)
        task.run()

        recon = task.get_data_to_cpu("object", as_numpy=True)[0]

        return recon

    @tutils.TungstenDataTester.wrap_recon_tester(
        name="test_2d_ptycho_epie_orthogonalize_opr_and_incoherent_modes"
    )
    def test_2d_ptycho_epie_orthogonalize_opr_and_incoherent_modes(self):
        self.setup_ptychi(cpu_only=False)

        data, probe, pixel_size_m, positions_px = self.load_tungsten_data(additional_opr_modes=3)

        options = api.EPIEOptions()
        options.data_options.data = data

        options.object_options.initial_guess = torch.ones(
            [1, *get_suggested_object_size(positions_px, probe.shape[-2:], extra=100)],
            dtype=get_default_complex_dtype(),
        )
        options.object_options.pixel_size_m = pixel_size_m
        options.object_options.optimizable = True
        options.object_options.optimizer = api.Optimizers.SGD
        options.object_options.step_size = 0.1
        options.object_options.alpha = 1

        options.probe_options.initial_guess = probe
        options.probe_options.optimizable = True
        options.probe_options.optimizer = api.Optimizers.SGD
        options.probe_options.step_size = 0.1
        options.probe_options.alpha = 1
        options.probe_options.eigenmode_update_relaxation = 0.1
        options.probe_options.orthogonalize_opr_modes.enabled = True
        options.probe_options.orthogonalize_incoherent_modes.enabled = True

        options.probe_position_options.position_x_px = positions_px[:, 1]
        options.probe_position_options.position_y_px = positions_px[:, 0]
        options.probe_position_options.optimizable = False

        options.opr_mode_weight_options.initial_weights = generate_initial_opr_mode_weights(
            len(positions_px), probe.shape[0]
        )
        options.opr_mode_weight_options.optimizable = True
        options.opr_mode_weight_options.update_relaxation = 0.1

        options.reconstructor_options.batch_size = 96
        options.reconstructor_options.num_epochs = 8
        options.reconstructor_options.allow_nondeterministic_algorithms = False

        task = PtychographyTask(options)
        task.run()

        recon = task.get_data_to_cpu("object", as_numpy=True)[0]

        return recon


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--generate-gold", action="store_true")
    parser.add_argument("--high-tol", action="store_true")
    args = parser.parse_args()

    tester = Test2dPtychoEpieOPR()
    tester.setup_method(name="", generate_data=False, generate_gold=args.generate_gold, debug=True)
    tester.test_2d_ptycho_epie_opr()
    tester.test_2d_ptycho_epie_orthogonalize_opr_and_incoherent_modes()



# Copyright © 2025 UChicago Argonne, LLC All right reserved
# Full license accessible at https://github.com//AdvancedPhotonSource/pty-chi/blob/main/LICENSE

# ruff: noqa: F401

try:
    from . import api
    from . import settings
    MOVIES_INSTALLED = True
except ImportError:
    MOVIES_INSTALLED = False


# Copyright © 2025 UChicago Argonne, LLC All right reserved
# Full license accessible at https://github.com//AdvancedPhotonSource/pty-chi/blob/main/LICENSE

import torch.nn as nn


class DoubleConv(nn.Module):
    """Double convolution layer comprised of
    ```(Conv2d -> BatchNorm -> ReLU) * 2```
    
    This implementation is adapted from https://github.com/milesial/Pytorch-UNet.
    """
    def __init__(self, in_channels, out_channels, mid_channels=None):
        super().__init__()
        if not mid_channels:
            mid_channels = out_channels
        self.double_conv = nn.Sequential(
            nn.Conv2d(in_channels, mid_channels, kernel_size=3, padding=1, bias=False),
            nn.BatchNorm2d(mid_channels),
            nn.ReLU(inplace=True),
            nn.Conv2d(mid_channels, out_channels, kernel_size=3, padding=1, bias=False),
            nn.BatchNorm2d(out_channels),
            nn.ReLU(inplace=True)
        )

    def forward(self, x):
        return self.double_conv(x)



# Copyright © 2025 UChicago Argonne, LLC All right reserved
# Full license accessible at https://github.com//AdvancedPhotonSource/pty-chi/blob/main/LICENSE

import os
import datetime
from typing import Union
import h5py
import numpy as np
import socket
import subprocess

# from ptychi.timer_utils import ADVANCED_TIME_DICT, ELAPSED_TIME_DICT, toggle_timer
import ptychi.timing.timer_utils as timer_utils


def get_timestamp_for_timing_files() -> str:
    current_datetime = datetime.datetime.now()
    timestamp = current_datetime.strftime("%Y-%m-%d_%H-%M-%S")
    date_string = current_datetime.strftime("%Y-%m-%d")
    time_string = current_datetime.strftime("%H-%M-%S")
    return timestamp, date_string, time_string


def save_timing_data_to_unique_file_path(name: str, folder: str):
    timing_results_folder = os.path.join(folder, name)
    if not os.path.exists(timing_results_folder):
        os.makedirs(timing_results_folder)

    if "TIMESTAMP" in os.environ:
        timestamp, date_string, time_string = (
            os.environ["TIMESTAMP"],
            os.environ["DATE_STRING"],
            os.environ["TIME_STRING"],
        )
    else:
        timestamp, date_string, time_string = get_timestamp_for_timing_files()

    unique_file_name = "timing_results_" + str(timestamp) + ".h5"
    file_path = os.path.join(timing_results_folder, unique_file_name)

    with h5py.File(file_path, "w") as F:
        F.attrs["name"] = name
        F.attrs["host"] = socket.gethostname().split(".")[0]
        F.attrs["commit"] = get_current_commit_hash()
        F.attrs["branch"] = get_current_branch_name()
        F.attrs["date"] = date_string
        F.attrs["time"] = time_string
        insert_timing_dict_into_h5_object(
            timer_utils.ELAPSED_TIME_DICT, F.create_group("elapsed_time_dict")
        )
        insert_timing_dict_into_h5_object(
            timer_utils.ADVANCED_TIME_DICT, F.create_group("advanced_time_dict")
        )


def read_timing_data_file(file_path: str) -> dict:
    def recursively_convert_h5_group(group):
        result = {}
        for key, item in group.items():
            if isinstance(item, h5py.Group):
                # Recursively process groups
                result[key] = recursively_convert_h5_group(item)
            elif isinstance(item, h5py.Dataset):
                # Convert datasets to numpy arrays or lists
                result[key] = item[()]  # Read the dataset
            else:
                raise ValueError(f"Unknown item type for key {key}: {type(item)}")
        return result

    with h5py.File(file_path, "r") as h5_file:
        d = recursively_convert_h5_group(h5_file)
        # Get attributes
        d["attributes"] = {}
        for k in h5_file.attrs.keys():
            d["attributes"][k] = h5_file.attrs[k]
        return d


def insert_timing_dict_into_h5_object(d: dict, h5_object: Union[h5py.Group, h5py.File]):
    for key, value in d.items():
        if isinstance(value, dict):
            # Recursively handle dicts
            insert_timing_dict_into_h5_object(value, h5_object.create_group(key))
        elif isinstance(value, np.ndarray):
            h5_object.create_dataset(key, data=value)
        else:
            raise ValueError("Data type not supported")


def get_current_commit_hash() -> str:
    # Get the shortened commit hash
    commit_hash = (
        subprocess.check_output(
            ["git", "rev-parse", "--short", "HEAD"],
            stderr=subprocess.STDOUT,
        )
        .strip()
        .decode("utf-8")
    )
    return commit_hash


def get_current_branch_name() -> str:
    # Get the current branch name
    branch_name = (
        subprocess.check_output(
            ["git", "rev-parse", "--abbrev-ref", "HEAD"],
            stderr=subprocess.STDOUT,
        )
        .strip()
        .decode("utf-8")
    )

    return branch_name



# Copyright © 2025 UChicago Argonne, LLC All right reserved
# Full license accessible at https://github.com//AdvancedPhotonSource/pty-chi/blob/main/LICENSE

from importlib.metadata import version, PackageNotFoundError

try:
    __version__ = version("ptychi")
except PackageNotFoundError:
    pass

from .device import Device, list_available_devices

__all__ = [
    "Device",
    "list_available_devices",
]



# Configuration file for Sphinx documentation

project = 'Pty-chi'
copyright = '2024'
author = 'Pty-chi Contributors'

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.napoleon',
    'sphinx.ext.viewcode',
    'sphinx.ext.intersphinx',
]

templates_path = ['_templates']
exclude_patterns = []

html_theme = 'sphinx_rtd_theme'
html_static_path = ['_static']

# Add any paths that contain custom static files
html_static_path = ['_static']

# Intersphinx mapping
intersphinx_mapping = {
    'python': ('https://docs.python.org/3', None),
    'numpy': ('https://numpy.org/doc/stable/', None),
    'torch': ('https://pytorch.org/docs/stable/', None),
} 


# Copyright © 2025 UChicago Argonne, LLC All right reserved
# Full license accessible at https://github.com//AdvancedPhotonSource/pty-chi/blob/main/LICENSE

import dataclasses

import ptychi.data_structures.base as dsbase
import ptychi.data_structures.object as object
import ptychi.data_structures.opr_mode_weights as oprweights
import ptychi.data_structures.probe as probe
import ptychi.data_structures.probe_positions as probepos


@dataclasses.dataclass
class ParameterGroup:
    def get_all_parameters(self) -> list["dsbase.ReconstructParameter"]:
        return list(self.__dict__.values())

    def get_optimizable_parameters(self) -> list["dsbase.ReconstructParameter"]:
        ovs = []
        for var in self.get_all_parameters():
            if var.optimizable:
                ovs.append(var)
        return ovs

    def get_config_dict(self):
        return {var.name: var.get_config_dict() for var in self.get_all_parameters()}


@dataclasses.dataclass
class PtychographyParameterGroup(ParameterGroup):
    object: "object.Object"

    probe: "probe.Probe"

    probe_positions: "probepos.ProbePositions"

    opr_mode_weights: "oprweights.OPRModeWeights"

    def __post_init__(self):
        if self.probe.has_multiple_opr_modes and self.opr_mode_weights is None:
            raise ValueError(
                "OPRModeWeights must be provided when the probe has multiple OPR modes."
            )


@dataclasses.dataclass
class PlanarPtychographyParameterGroup(PtychographyParameterGroup):
    object: "object.PlanarObject"


# Copyright © 2025 UChicago Argonne, LLC All right reserved
# Full license accessible at https://github.com//AdvancedPhotonSource/pty-chi/blob/main/LICENSE

from typing import Optional, TYPE_CHECKING

import torch
from torch.utils.data import Dataset

from ptychi.reconstructors.ad_general import AutodiffReconstructor
from ptychi.reconstructors.base import IterativePtychographyReconstructor
import ptychi.metrics as metrics

if TYPE_CHECKING:
    import ptychi.data_structures.parameter_group as pg
    import ptychi.api as api


class AutodiffPtychographyReconstructor(AutodiffReconstructor, IterativePtychographyReconstructor):
    def __init__(
        self,
        parameter_group: "pg.PtychographyParameterGroup",
        dataset: Dataset,
        options: Optional["api.options.ad_ptychography.AutodiffPtychographyOptions"] = None,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(
            parameter_group=parameter_group,
            dataset=dataset,
            options=options,
            *args,
            **kwargs,
        )
    
    def build(self):
        self.update_requires_grad()
        super().build()
        
    def build_forward_model(self):
        self.forward_model_params["wavelength_m"] = self.dataset.wavelength_m
        self.forward_model_params["detector_size"] = tuple(self.dataset.patterns.shape[-2:])
        self.forward_model_params["free_space_propagation_distance_m"] = self.dataset.free_space_propagation_distance_m
        self.forward_model_params["pad_for_shift"] = self.options.forward_model_options.pad_for_shift
        self.forward_model_params["low_memory_mode"] = self.options.forward_model_options.low_memory_mode
        return super().build_forward_model()
    
    def run_pre_epoch_hooks(self) -> None:
        super().run_pre_epoch_hooks()
        self.update_requires_grad()
                    
    def update_requires_grad(self):
        """Set requires_grad of variables that don't need to be optimized at this epoch to False.
        This prevents these variables from holding the old graph and causing errors during
        backward() call.
        """
        for var in self.parameter_group.get_all_parameters():
            if not (var.options.optimization_plan.is_enabled(self.current_epoch) and var.options.optimizable):
                var.tensor.requires_grad = False
            else:
                var.tensor.requires_grad = True
            for subvar in var.sub_modules:
                if not (subvar.options.optimization_plan.is_enabled(self.current_epoch) and subvar.options.optimizable):
                    subvar.tensor.requires_grad = False
                else:
                    subvar.tensor.requires_grad = True

    def run_post_differentiation_hooks(self, input_data, y_true):
        super().run_post_differentiation_hooks(input_data, y_true)

        # If OPRModeWeights is optimized in the current epoch (i.e., it has gradient) but intensity variation
        # optimization is not enabled, set the gradient of the principal mode's weights to 0. Similar is done
        # for the gradient of the eigenmode weights.
        if self.parameter_group.opr_mode_weights.optimization_enabled(self.current_epoch):
            if not self.parameter_group.opr_mode_weights.intensity_variation_optimization_enabled(
                self.current_epoch
            ):
                self.parameter_group.opr_mode_weights.tensor.grad[:, 0] = 0
            if not self.parameter_group.opr_mode_weights.eigenmode_weight_optimization_enabled(
                self.current_epoch
            ):
                self.parameter_group.opr_mode_weights.tensor.grad[:, 1:] = 0

    def run_post_update_hooks(self) -> None:
        with torch.no_grad():
            if self.parameter_group.object.optimization_enabled(self.current_epoch):
                self.parameter_group.object.post_update_hook()

            if self.parameter_group.probe.optimization_enabled(self.current_epoch):
                self.parameter_group.probe.post_update_hook()

            if self.parameter_group.probe_positions.optimization_enabled(self.current_epoch):
                self.parameter_group.probe_positions.post_update_hook()

    def apply_regularizers(self) -> None:
        """
        Apply Tikonov regularizers, e.g., L1 norm for the object.

        This function calculates the regularization terms and backpropagates them. Since the gradients
        of the parameters haven't been zeroed, the regularizers' gradients will be added to them.
        """
        reg = super().apply_regularizers()

        object_ = self.parameter_group.object
        if object_.options.l1_norm_constraint.is_enabled_on_this_epoch(self.current_epoch):
            # object.data returns a copy, so we directly access the tensor here.
            obj_data = object_.tensor.data[..., 0] + 1j * object_.tensor.data[..., 1]
            reg = reg + object_.options.l1_norm_constraint.weight * torch.mean(obj_data.abs())
            
        if object_.options.l2_norm_constraint.is_enabled_on_this_epoch(self.current_epoch):
            obj_data = object_.tensor.data[..., 0] + 1j * object_.tensor.data[..., 1]
            reg = reg + object_.options.l2_norm_constraint.weight * torch.mean(obj_data.abs() ** 2)
            
        if object_.options.total_variation.is_enabled_on_this_epoch(self.current_epoch):
            obj_data = object_.tensor.data[..., 0] + 1j * object_.tensor.data[..., 1]
            reg = reg + object_.options.total_variation.weight * metrics.TotalVariationLoss(reduction="mean")(obj_data)
            
        if reg.requires_grad:
            reg.backward()
        return reg
    
    def get_retain_graph(self) -> bool:
        """Determines whether `loss.backward()` should have `retain_graph=True`.
        Since regularizers' gradients are computed after `loss.backward()`, we need to
        return `True` if any regularizer is enabled.
        """
        if self.parameter_group.object.options.experimental.deep_image_prior_options.enabled:
            return True
        if self.parameter_group.probe.options.experimental.deep_image_prior_options.enabled:
            return True
        if (
            self.parameter_group.object.options.l1_norm_constraint.is_enabled_on_this_epoch(self.current_epoch)
            and self.parameter_group.object.options.l2_norm_constraint.is_enabled_on_this_epoch(self.current_epoch)
            and self.parameter_group.object.options.total_variation.is_enabled_on_this_epoch(self.current_epoch)
        ):
            return True
        return False

    def run_minibatch(self, input_data, y_true, *args, **kwargs):        
        y_pred = self.forward_model(*input_data)
        batch_loss = self.loss_function(
            y_pred[:, self.dataset.valid_pixel_mask], y_true[:, self.dataset.valid_pixel_mask]
        )

        batch_loss.backward(retain_graph=self.get_retain_graph())
        self.run_post_differentiation_hooks(input_data, y_true)
        reg_loss = self.apply_regularizers()
        
        self.step_all_optimizers()
        self.forward_model.zero_grad()
        self.run_post_update_hooks()
        
        self.loss_tracker.update_batch_loss(y_pred=y_pred, y_true=y_true, loss=batch_loss.item())
        self.loss_tracker.update_batch_regularization_loss(reg_loss.item())



# Copyright © 2025 UChicago Argonne, LLC All right reserved
# Full license accessible at https://github.com//AdvancedPhotonSource/pty-chi/blob/main/LICENSE

from typing import Union, Optional
import dataclasses

from numpy import ndarray, inf
from torch import Tensor

import ptychi.api.options.base as base


__all__ = ["PtychographyDataOptions"]


@dataclasses.dataclass
class PtychographyDataOptions(base.Options):

    data: Union[ndarray, Tensor] = None
    """
    The intensity data. Use collected data as they are; data should NOT be FFT-shifted 
    or square-rooted.
    """

    free_space_propagation_distance_m: float = inf
    """The free-space propagation distance in meters, or `inf` for far-field."""

    wavelength_m: float = 1e-9
    """The wavelength in meters."""

    fft_shift: bool = True
    """Whether to FFT-shift the diffraction data."""

    detector_pixel_size_m: float = 1e-8
    """The detector pixel size in meters."""

    valid_pixel_mask: Optional[Union[ndarray, Tensor]] = None
    """A 2D boolean mask where valid pixels are True."""
    
    save_data_on_device: bool = False
    """Whether to save the diffraction data on acceleration devices like GPU."""



# Copyright © 2025 UChicago Argonne, LLC All right reserved
# Full license accessible at https://github.com//AdvancedPhotonSource/pty-chi/blob/main/LICENSE

from typing import Optional
import numpy as np

from .settings import MovieFileSettings, MovieFileTypes

from .mappings import prepare_movie_subject, movie_setting_types
from .io import append_array_to_h5, save_movie_to_file
import os
import h5py

import ptychi.reconstructors.base as base

dataset_name = "frames"


class MovieBuilder:
    current_frame: int = 0
    frames: np.ndarray = None
    epochs: list[int] = []

    def __init__(
        self,
        settings: movie_setting_types,
        folder: str,
        movie_name: str,
    ):
        self.settings = settings
        self.folder = folder
        self.movie_name = movie_name
        if self.settings.save_intermediate_data_to_hdf5:
            self.hdf5_file_path = os.path.join(self.folder, self.movie_name + ".hdf5")
        else:
            self.hdf5_file_path = None

        if not os.path.exists(self.folder):
            os.mkdir(self.folder)

    def save_frame_enabled_this_epoch(self, current_epoch: int):
        if current_epoch % self.settings.snapshot.stride == 0:
            return True
        else:
            return False

    def record_frame(self, movie_frame_array: np.ndarray):
        if self.settings.save_intermediate_data_to_hdf5:
            append_array_to_h5(
                array=movie_frame_array,
                file_path=self.hdf5_file_path,
                create_new_file=self.current_frame == 0,
            )
        else:
            if self.current_frame == 0:
                self.frames = movie_frame_array[None]
            else:
                self.frames = np.append(self.frames, movie_frame_array[None], axis=0)
        self.epochs += [self.current_frame * self.settings.snapshot.stride]
        self.current_frame += 1

    def create_movie(
        self, file_type: MovieFileTypes, movie_file_settings: Optional[MovieFileSettings] = None
    ):
        if movie_file_settings is None:
            movie_file_settings = self.settings.movie_file
        output_path = os.path.join(self.folder, self.movie_name + "." + file_type)
        if self.settings.save_intermediate_data_to_hdf5:
            with h5py.File(self.hdf5_file_path) as F:
                frames = F[dataset_name][:]
        else:
            frames = self.frames
        save_movie_to_file(
            array=frames,
            file_type=file_type,
            output_path=output_path,
            fps=self.settings.movie_file.fps,
            enhance_contrast=self.settings.movie_file.high_contrast,
            colormap=self.settings.movie_file.colormap,
            titles=[str(i) for i in self.epochs],
            compress=self.settings.movie_file.compress,
            upper_bound=self.settings.movie_file.upper_bound,
            lower_bound=self.settings.movie_file.lower_bound,
        )

    def reset(self):
        self.current_frame = 0
        self.epochs = []
        self.frames = None


class MoviesManager:
    """Class that handles the updating of `MovieBuilder` instances."""

    movie_builders: dict[str, MovieBuilder] = {}

    def add_movie_builder(
        self,
        settings: movie_setting_types,
        folder: str,
        movie_name: str,
    ):
        self.movie_builders[movie_name] = MovieBuilder(settings, folder, movie_name)

    def record_all_frames(self, reconstructor: "base.Reconstructor"):
        for movie_builder in self.movie_builders.values():
            if movie_builder.save_frame_enabled_this_epoch(reconstructor.current_epoch):
                movie_frame_array = prepare_movie_subject(reconstructor, movie_builder.settings)
                movie_builder.record_frame(movie_frame_array)

    def create_all_movies(self, file_type: MovieFileTypes):
        for movie_builder in self.movie_builders.values():
            movie_builder.create_movie(file_type)

    def reset_all(self):
        for movie_builder in self.movie_builders.values():
            movie_builder.reset()

    def delete_intermediate_movie_files(self):
        for movie_builder in self.movie_builders.values():
            file_path = movie_builder.hdf5_file_path
            if file_path is not None and os.path.exists(file_path):
                os.remove(file_path)
                print(f"Intermediate file {os.path.basename(file_path)} has been deleted.")



import ptychi.movies as movies
if not movies.MOVIES_INSTALLED:
    raise ImportError("Dependencies needed by ptychi.movies are not installed!")

import sys
import os
import torch
import numpy as np
import cv2
import ptychi.api as api
from ptychi.api.task import PtychographyTask
from ptychi.utils import get_suggested_object_size, get_default_complex_dtype

# Add the folder containing "tests/" to the path
sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))
from tests.test_utils import TungstenDataTester

from ptychi.movies.settings import (
    ProbePlotTypes,
    ProcessFunctionType,
    ProbeMovieSettings,
    ObjectMovieSettings,
    MovieFileTypes,
    SnapshotSettings,
    MovieFileSettings,
)

movie_folder = input("Enter a folder to save movies to:\n")

### Movie setup ####
# Enable movies
movies.api.toggle_movies(True)
### Initialize movie builders ###
# The scale determines the level of downsampling
scale = 1
# 1) Add movie builder for total probe intensity
probe_sum_movie_name = "probe_total_intensity"
movies.api.add_new_movie_builder(
    settings=ProbeMovieSettings(snapshot=SnapshotSettings(scale=scale)),
    folder=movie_folder,
    movie_name=probe_sum_movie_name,
)
# You can change settings after initializing the MovieBuilder object by
# using the get_movie_builder_settings function to retrieve the settings
probe_movie_file_settings = movies.api.get_movie_builder_settings(probe_sum_movie_name).movie_file
probe_movie_file_settings.colormap = cv2.COLORMAP_VIRIDIS

# 2) Add movie builder for showing all probe mode magnitudes
probe_modes_movie_name = "probe_modes_magnitude"
movies.api.add_new_movie_builder(
    settings=ProbeMovieSettings(
        plot_type=ProbePlotTypes.SEPERATE_MODES,
        snapshot=SnapshotSettings(scale=scale),
        movie_file=MovieFileSettings(colormap=cv2.COLORMAP_VIRIDIS),
    ),
    folder=movie_folder,
    movie_name=probe_modes_movie_name,
)

# 3) Add movie builder for showing all probe mode phases
probe_modes_movie_name = "probe_modes_phase"
movies.api.add_new_movie_builder(
    settings=ProbeMovieSettings(
        process_function=ProcessFunctionType.PHASE,
        plot_type=ProbePlotTypes.SEPERATE_MODES,
        snapshot=SnapshotSettings(scale=scale),
        movie_file=MovieFileSettings(colormap=cv2.COLORMAP_JET),
    ),
    folder=movie_folder,
    movie_name=probe_modes_movie_name,
)

# 4) Add movie builder for object phase (this is the default for object movie settings)
movies.api.add_new_movie_builder(
    settings=ObjectMovieSettings(snapshot=SnapshotSettings(scale=scale)),
    folder=movie_folder,
    movie_name="reconstructed_object_phase",
)

# 5) Add movie builder for object magnitude
# Enabling save_intermediate_data_to_hdf5 means that each frame of the
# movie will *not* be saved to a variable in python, instead it will be
# appended to a dataset in a hdf5 each iteration. This is useful if you
# know the array of saved frames will be very large and therefore would take
# up a lot of memory.
movies.api.add_new_movie_builder(
    settings=ObjectMovieSettings(
        process_function=ProcessFunctionType.MAGNITUDE,
        snapshot=SnapshotSettings(scale=scale),
        save_intermediate_data_to_hdf5=True,
    ),
    folder=movie_folder,
    movie_name="reconstructed_object_magnitude",
)

### Run ptychography reconstruction ###
demo_tester = TungstenDataTester()
demo_tester.load_tungsten_data()
data, probe, pixel_size_m, positions_px = demo_tester.load_tungsten_data(additional_opr_modes=0)

options = api.DMOptions()
options.data_options.data = data

options.object_options.initial_guess = torch.ones(
    [1, *get_suggested_object_size(positions_px, probe.shape[-2:], extra=100)],
    dtype=get_default_complex_dtype(),
)
options.object_options.pixel_size_m = pixel_size_m
options.object_options.optimizable = True

options.probe_options.initial_guess = probe
options.probe_options.power_constraint.probe_power = np.sum(np.max(data, axis=-3), axis=(-2, -1))
options.probe_options.power_constraint.enabled = True
options.probe_options.optimizable = True

options.probe_position_options.position_x_px = positions_px[:, 1]
options.probe_position_options.position_y_px = positions_px[:, 0]
options.probe_position_options.optimizable = True

options.reconstructor_options.num_epochs = 20
options.reconstructor_options.chunk_length = 100

task = PtychographyTask(options)
task.run()


### Generate movie files ###
movies.api.create_all_movies(MovieFileTypes.GIF)
# You don't need to use the StrEnum, you could just use: movies.api.create_all_movies("gif")
# Check MovieFileTypes to see what file formats are available

### Delete intermediate movie files after generating movies ###
movies.api.delete_intermediate_movie_files()



import argparse

import torch

import ptychi.image_proc as ip
import test_utils as tutils


class TestPatchExtractionPlacement(tutils.BaseTester):
    
    def test_patch_extraction_placement_set_ordinary(self):
        torch.manual_seed(123)
        
        positions = torch.rand(10, 2) * 64 + 32
        image = torch.rand(128, 128)
        
        patches = ip.extract_patches_fourier_shift(image, positions, (32, 32))
        image_replaced = ip.place_patches_fourier_shift(torch.zeros_like(image), positions, patches, op="set", adjoint_mode=False)
        
        mask = image_replaced > 0
        orig_sum = torch.sum(image[mask])
        replaced_sum = torch.sum(image_replaced[mask])
        
        print("Mode: set")
        print("Sum of original image ROI:", orig_sum)
        print("Sum of replaced image ROI:", replaced_sum)
        print("Max of absolute difference:", torch.max(torch.abs(image_replaced[mask] - image[mask])))
        
        if self.debug:
            import matplotlib.pyplot as plt
            fig, ax = plt.subplots(1, 2)
            ax[0].imshow(image)
            ax[1].imshow(image_replaced)
            plt.suptitle("Mode: set")
            plt.tight_layout()
            plt.show()
        
        assert torch.allclose(orig_sum, replaced_sum, rtol=1e-2)
        
    def test_patch_extraction_placement_add_adjoint(self):
        torch.manual_seed(123)
        
        positions = torch.tensor([[64, 64]])
        image = torch.zeros([128, 128])
        nonzero_rad = 15
        image[64 - nonzero_rad:64 + nonzero_rad, 64 - nonzero_rad:64 + nonzero_rad] = torch.rand(2 * nonzero_rad, 2 * nonzero_rad)
        
        patches = ip.extract_patches_fourier_shift(image, positions, (32, 32))
        image_replaced = ip.place_patches_fourier_shift(torch.zeros_like(image), positions, patches, op="add", adjoint_mode=True)
                
        if self.debug:
            import matplotlib.pyplot as plt
            fig, ax = plt.subplots(1, 2)
            ax[0].imshow(image)
            ax[1].imshow(image_replaced)
            plt.suptitle("Mode: add")
            plt.tight_layout()
            plt.show()
        
        # Verify if extraction and placement are Hermitian adjoints using the identity 
        #   <Ax, y> = <x, A^H y>
        # where A is the extraction operator, A^H is the placement operator, 
        # x is the original image, and y is the extracted patch; y = Ax.
        ax_y = torch.sum(patches.abs() ** 2)
        x_ahy = torch.sum(image_replaced * image)
        assert torch.allclose(ax_y, x_ahy)
        
    def test_patch_extraction_placement_int(self):        
        positions = torch.tensor([[5, 5], [20, 20], [15, 20]]) + 0.5
        image = torch.arange(32 * 32).reshape(32, 32)
        
        patches = ip.extract_patches_integer(image, positions, (8, 8))
        image_replaced = ip.place_patches_integer(torch.zeros_like(image), positions, patches, op="add")
                
        ax_y = torch.sum(patches.abs() ** 2)
        x_ahy = torch.sum(image_replaced * image)
        assert torch.allclose(ax_y, x_ahy)
        assert torch.all(patches.sum(-1).sum(-1) == torch.tensor([11616, 43296, 33056]))


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--generate-gold', action='store_true')
    args = parser.parse_args()
    
    tester = TestPatchExtractionPlacement()
    tester.setup_method(name="", generate_data=False, generate_gold=args.generate_gold, debug=True)
    tester.test_patch_extraction_placement_set_ordinary()
    tester.test_patch_extraction_placement_add_adjoint()
    tester.test_patch_extraction_placement_int()



# Copyright © 2025 UChicago Argonne, LLC All right reserved
# Full license accessible at https://github.com//AdvancedPhotonSource/pty-chi/blob/main/LICENSE

# ruff: noqa: F403

from .options import *
from .enums import *



# Copyright © 2025 UChicago Argonne, LLC All right reserved
# Full license accessible at https://github.com//AdvancedPhotonSource/pty-chi/blob/main/LICENSE

from typing import Optional

from dataclasses import dataclass


@dataclass
class OptimizationPlan:
    """
    When a `ReconstructParameter` has `optimizable == True`, this class is used to specify
    the start, stop, and stride epochs of the optimization for that parameter. This class is
    also used by `FeatureOptions`.
    """
    start: int = 0
    """
    The starting epoch.
    """
    
    stop: Optional[int] = None
    """
    The starting epoch. If None, optimization will run to the last epoch if the parameter
    is optimizable.
    """
    
    stride: int = 1
    """
    The stride in epochs. Optimization will run every `stride` epochs.
    """
    
    def is_enabled(self, epoch: int) -> bool:
        if self.start is not None and epoch < self.start:
            return False
        if self.stop is not None and epoch >= self.stop:
            return False
        if self.start is None:
            return True
        return (epoch - self.start) % self.stride == 0
    
    def is_in_optimization_interval(self, epoch: int) -> bool:
        if self.start is not None and epoch < self.start:
            return False
        if self.stop is not None and epoch >= self.stop:
            return False
        return True
    


# Copyright © 2025 UChicago Argonne, LLC All right reserved
# Full license accessible at https://github.com//AdvancedPhotonSource/pty-chi/blob/main/LICENSE

from collections.abc import Sequence
from dataclasses import dataclass

import torch


@dataclass(frozen=True)
class Device:
    backend: str
    ordinal: int
    name: str

    @property
    def torch_device(self) -> str:
        return f"{self.backend.lower()}:{self.ordinal}"

def list_available_devices() -> Sequence[Device]:
    available_devices = list()

    if torch.cuda.is_available():
        for ordinal in range(torch.cuda.device_count()):
            name = torch.cuda.get_device_name(ordinal)
            device = Device("cuda", ordinal, name)
            available_devices.append(device)

    if torch.xpu.is_available():
        for ordinal in range(torch.xpu.device_count()):
            name = torch.xpu.get_device_name(ordinal)
            device = Device("xpu", ordinal, name)
            available_devices.append(device)

    return available_devices



# Copyright © 2025 UChicago Argonne, LLC All right reserved
# Full license accessible at https://github.com//AdvancedPhotonSource/pty-chi/blob/main/LICENSE

from typing import Union, TYPE_CHECKING, Optional
import logging

import torch
from torch import Tensor

import ptychi.data_structures.base as dsbase
import ptychi.image_proc as ip
import ptychi.maths as pmath
from ptychi.timing.timer_utils import timer

if TYPE_CHECKING:
    import ptychi.api as api
    from ptychi.data_structures.probe import Probe

logger = logging.getLogger(__name__)


class OPRModeWeights(dsbase.ReconstructParameter):
    # TODO: update_relaxation is only used for LSQML. We should create dataclasses
    # to contain additional options for ReconstructParameter classes, and subclass them for specific
    # reconstruction algorithms - for example, OPRModeWeightsOptions -> LSQMLOPRModeWeightsOptions.
    options: "api.options.base.OPRModeWeightsOptions"

    def __init__(
        self,
        *args,
        name="opr_weights",
        options: "api.options.base.OPRModeWeightsOptions" = None,
        **kwargs,
    ):
        """
        Weights of OPR modes for each scan point.

        Parameters
        ----------
        name : str
            Name of the parameter.
        update_relaxation : float
            Relaxation factor, or effectively the step size, for the update step in LSQML.
        """
        super().__init__(*args, name=name, options=options, is_complex=False, **kwargs)
        if len(self.shape) != 2:
            raise ValueError("OPR weights must be of shape (n_scan_points, n_opr_modes).")

        if self.optimizable:
            if not (options.optimize_eigenmode_weights or options.optimize_intensity_variation):
                raise ValueError(
                    "When OPRModeWeights is optimizable, at least 1 of "
                    "optimize_eigenmode_weights and optimize_intensity_variation "
                    "should be set to True."
                )

        self.optimize_eigenmode_weights = options.optimize_eigenmode_weights
        self.optimize_intensity_variation = options.optimize_intensity_variation

    @property
    def n_opr_modes(self):
        return self.tensor.shape[1]

    @property
    def n_scan_points(self):
        return self.tensor.shape[0]

    @property
    def n_eigenmodes(self):
        return self.tensor.shape[1] - 1

    def build_optimizer(self):
        if self.optimizer_class is None:
            return
        if self.optimizable:
            if isinstance(self.tensor, dsbase.ComplexTensor):
                self.optimizer = self.optimizer_class([self.tensor.data], **self.optimizer_params)
            else:
                self.optimizer = self.optimizer_class([self.tensor], **self.optimizer_params)

    def get_weights(self, indices: Union[tuple[int, ...], slice]) -> Tensor:
        return self.data[indices]

    def optimization_enabled(self, epoch: int):
        enabled = super().optimization_enabled(epoch)
        return enabled and (self.optimize_eigenmode_weights or self.optimize_intensity_variation)

    def eigenmode_weight_optimization_enabled(self, epoch: int):
        enabled = super().optimization_enabled(epoch)
        return enabled and self.optimize_eigenmode_weights

    def intensity_variation_optimization_enabled(self, epoch: int):
        enabled = super().optimization_enabled(epoch)
        return enabled and self.optimize_intensity_variation

    @timer()
    def update_variable_probe(
        self,
        probe: "Probe",
        indices: Tensor,
        chi: Tensor,
        delta_p_i: Tensor,
        delta_p_hat: Tensor,
        obj_patches: Tensor,
        current_epoch: int,
        probe_mode_index: Optional[int] = None,
    ):
        # TODO: OPR updates are calculated using chi with uniquely shifted probes, 
        # probe without shift, and probe updates that are adjoint-shifted. This is
        # not accurate, but PtychoShelves does the same. We might need to revisit this.
        mode_slicer = probe._get_probe_mode_slicer(probe_mode_index)
        chi = chi[:, mode_slicer]
        delta_p_i = delta_p_i[:, mode_slicer]
        delta_p_hat = delta_p_hat[mode_slicer]

        if probe.has_multiple_opr_modes and (
            probe.optimization_enabled(current_epoch)
            or (self.eigenmode_weight_optimization_enabled(current_epoch))
        ):
            self.update_opr_probe_modes_and_weights(
                probe, indices, chi, delta_p_i, delta_p_hat, obj_patches, current_epoch
            )

        if self.intensity_variation_optimization_enabled(current_epoch):
            delta_weights_int = self._calculate_intensity_variation_update_direction(
                probe,
                indices,
                chi,
                obj_patches,
            )
            self._apply_variable_intensity_updates(delta_weights_int)

    @timer()
    def update_opr_probe_modes_and_weights(
        self,
        probe: "Probe",
        indices: Tensor,
        chi: Tensor,
        delta_p_i: Tensor,
        delta_p_hat: Tensor,
        obj_patches: Tensor,
        current_epoch: int,
    ):
        """
        Update the eigenmodes of the first incoherent mode of the probe, and update the OPR mode weights.

        This implementation is adapted from PtychoShelves code (update_variable_probe.m) and has some
        differences from Eq. 31 of Odstrcil (2018).
        """
        probe_data = probe.data
        weights_data = self.data

        batch_size = len(delta_p_i)
        n_points_total = self.n_scan_points

        # If there is only one sample in the batch, `residue_update` would become 0
        # which causes division by zero error. We skip the update if that's the case.
        if batch_size == 1:
            return

        # FIXME: reduced relax_u/v by a factor of 10 for stability, but PtychoShelves works without this.
        relax_u = min(0.1, batch_size / n_points_total) * probe.options.eigenmode_update_relaxation
        relax_v = self.options.update_relaxation
        # Shape of delta_p_i:       (batch_size, n_probe_modes, h, w)
        # Use only the first incoherent mode
        delta_p_i = delta_p_i[:, 0, :, :]
        delta_p_hat = delta_p_hat[0, :, :]
        residue_update = delta_p_i - delta_p_hat

        # Start from the second OPR mode which is the first after the main mode - i.e., the first eigenmode.
        for i_opr_mode in range(1, probe.n_opr_modes):
            # Just take the first incoherent mode.
            eigenmode_i = probe.get_mode_and_opr_mode(mode=0, opr_mode=i_opr_mode)
            weights_i = self.get_weights(indices)[:, i_opr_mode]
            eigenmode_i, weights_i = self._update_first_eigenmode_and_weight(
                residue_update,
                eigenmode_i,
                weights_i,
                relax_u,
                relax_v,
                obj_patches,
                chi,
                update_eigenmode=probe.optimization_enabled(current_epoch),
                update_weights=self.eigenmode_weight_optimization_enabled(current_epoch),
            )

            # Project residue on this eigenmode, then subtract it.
            if i_opr_mode < probe.n_opr_modes - 1:
                residue_update = residue_update - pmath.project(
                    residue_update, eigenmode_i, dim=(-2, -1)
                )

            probe_data[i_opr_mode, 0, :, :] = eigenmode_i
            weights_data[indices, i_opr_mode] = weights_i

        if probe.optimization_enabled(current_epoch):
            probe.set_data(probe_data)
        if self.eigenmode_weight_optimization_enabled(current_epoch):
            self.set_data(weights_data)

    @timer()
    def _update_first_eigenmode_and_weight(
        self,
        residue_update: Tensor,
        eigenmode_i: Tensor,
        weights_i: Tensor,
        relax_u: Tensor,
        relax_v: Tensor,
        obj_patches: Tensor,
        chi: Tensor,
        eps=1e-5,
        update_eigenmode=True,
        update_weights=True,
    ):
        # Shape of residue_update:          (batch_size, h, w)
        # Shape of eigenmode_i:             (h, w)
        # Shape of weights_i:               (batch_size,)

        obj_patches = obj_patches[:, 0]

        # Update eigenmode.
        # Shape of proj:                    (batch_size, h, w)
        # FIXME: What happens when weights is zero!?
        proj = ((residue_update.conj() * eigenmode_i).real + weights_i[:, None, None]) / pmath.norm(
            weights_i
        ) ** 2

        if update_eigenmode:
            # Shape of eigenmode_update:        (h, w)
            eigenmode_update = torch.mean(
                residue_update * torch.mean(proj, dim=(-2, -1), keepdim=True), dim=0
            )
            eigenmode_i = eigenmode_i + relax_u * eigenmode_update / (
                pmath.mnorm(eigenmode_update.view(-1))
            )
            eigenmode_i = eigenmode_i / pmath.mnorm(eigenmode_i.view(-1))

        if update_weights:
            # Update weights using Eq. 23a.
            # Shape of psi:                     (batch_size, h, w)
            psi = eigenmode_i * obj_patches
            # The denominator can get smaller and smaller as eigenmode_i goes down.
            # Weight update needs to be clamped.
            denom = torch.mean((torch.abs(psi) ** 2), dim=(-2, -1))
            num = torch.mean((chi[:, 0, :, :] * psi.conj()).real, dim=(-2, -1))
            weight_update = num / (denom + 0.1 * torch.mean(denom))
            # weight_update = weight_update.clamp(max=10)
            weights_i = weights_i + relax_v * weight_update

        return eigenmode_i, weights_i

    @timer()
    def _calculate_intensity_variation_update_direction(
        self,
        probe: "Probe",
        indices: Tensor,
        chi: Tensor,
        obj_patches: Tensor,
    ):
        """
        Update variable intensity scaler - i.e., the OPR mode weight corresponding to the main mode.

        This implementation is adapted from PtychoShelves code (update_variable_probe.m) and has some
        differences from Eq. 31 of Odstrcil (2018).
        """
        # Just take the first slice.
        obj_patches = obj_patches[:, 0]

        mean_probe = probe.get_mode_and_opr_mode(mode=0, opr_mode=0)
        op = obj_patches * mean_probe
        num = torch.real(op.conj() * chi[:, 0, ...])
        denom = op.abs() ** 2
        delta_weights_int_i = torch.sum(num, dim=(-2, -1)) / torch.sum(denom, dim=(-2, -1))
        # Pad it to the same shape as opr_mode_weights.
        delta_weights_int = torch.zeros_like(self.data)
        delta_weights_int[indices, 0] = delta_weights_int_i
        return delta_weights_int

    @timer()
    def _apply_variable_intensity_updates(self, delta_weights_int: Tensor):
        self.set_data(self.data + 0.1 * delta_weights_int)

    @timer()
    def smooth_weights(self):
        """
        Smooth the weights with a median filter.
        """
        if self.options.smoothing.method is None:
            return
        weights = self.data
        if self.options.smoothing.method == "median":
            if self.n_scan_points < 81:
                logger.warning(
                    "OPR weight smoothing with median filter could "
                    "not run because the number of scan points is less than 81."
                )
                return
            weights = ip.median_filter_1d(weights.T, window_size=81).T
        elif self.options.smoothing.method == "polynomial":
            if self.n_scan_points < self.options.smoothing.polynomial_degree:
                logger.warning(
                    "OPR weight smoothing with polynomial filter could "
                    "not run because the number of scan points is less than the "
                    "polynomial smoothing degree ({}).".format(
                        self.options.smoothing.polynomial_degree
                    )
                )
                return
            inds = torch.arange(self.n_scan_points, device=weights.device, dtype=weights.dtype)
            for i_opr_mode in range(1, self.n_opr_modes):
                weights_current_mode = weights[:, i_opr_mode]
                fit_coeffs = pmath.polyfit(
                    inds, weights_current_mode, deg=self.options.smoothing.polynomial_degree
                )
                weights_smoothed = pmath.polyval(inds, fit_coeffs)
                weights[:, i_opr_mode] = 0.5 * weights_current_mode + 0.5 * weights_smoothed
        self.set_data(weights)

    @timer()
    def remove_outliers(self):
        aevol = torch.abs(self.data)
        weights = torch.min(aevol, 1.5 * torch.quantile(aevol, 0.95)) * torch.sign(self.data)
        self.set_data(weights)



# Copyright © 2025 UChicago Argonne, LLC All right reserved
# Full license accessible at https://github.com//AdvancedPhotonSource/pty-chi/blob/main/LICENSE

from typing import TypeAlias, Any
import numpy
import torch


BooleanArray: TypeAlias = numpy.typing.NDArray[numpy.bool_]
IntegerArray: TypeAlias = numpy.typing.NDArray[numpy.integer[Any]]
RealArray: TypeAlias = numpy.typing.NDArray[numpy.floating[Any]]
ComplexArray: TypeAlias = numpy.typing.NDArray[numpy.complexfloating[Any, Any]]

BooleanTensor: TypeAlias = torch.Tensor
ComplexTensor: TypeAlias = torch.Tensor
RealTensor: TypeAlias = torch.Tensor
